from gpim import gprutils as utils
from gpim.gpreg.gpr import reconstructor
from gpim.gpreg.skgpr import skreconstructor
from gpim.gpreg.vgpr import vreconstructor
from gpim.gpbayes.boptim import boptimizer



'''
acqfunc.py
======
Acquisition functions
'''

import numpy as np
from scipy.stats import norm


def confidence_bound(gpmodel, X_full, **kwargs):
    """
    Confidence bound acquisition function
    (a modification of upper confidence bound)

    Args:
        gpmodel (gpim reconstructor object):
            Surrogate function that allows computing
            mean and standard deviation
        X_full (ndarray):
            Full grid indices
        **alpha (float):
            :math:`\\alpha` coefficient in :math:`\\alpha \\mu + \\beta \\sigma`
        **beta (float):
            :math:`\\beta` coefficient in :math:`\\alpha \\mu + \\beta \\sigma`

    Returns:
        Acquisition function values and GP prediction (mean + stddev)
    """
    alpha = kwargs.get("alpha", 0)
    beta = kwargs.get("beta", 1)
    mean, sd = gpmodel.predict(X_full, verbose=0)
    acq = alpha * mean + beta * sd
    return acq, (mean, sd)


def expected_improvement(gpmodel, X_full, X_sparse, **kwargs):
    """
    Expected improvement acquisition function

    Args:
        gpmodel (gpim reconstructor object):
            Surrogate function that allows computing
            mean and standard deviation
        X_full (ndarray):
            Full grid indices
        X_sparse (ndarray):
            Sparse grid indices
        **xi (float):
            xi constant value

    Returns:
        Acquisition function values and GP prediction (mean + stddev)
    """
    xi = kwargs.get("xi", 0.01)
    mean, sd = gpmodel.predict(X_full, verbose=0)
    mean_sample, _ = gpmodel.predict(X_sparse, verbose=0)

    mean_sample_opt = np.nanmax(mean_sample)
    imp = mean - mean_sample_opt - xi
    z = imp / sd
    acq = imp * norm.cdf(z) + sd * norm.pdf(z)
    return acq, (mean, sd)


def probability_of_improvement(gpmodel, X_full, X_sparse, **kwargs):
    """
    Probability of improvement acquisition function

    Args:
        gpmodel (gpim reconstructor object):
            Surrogate function that allows computing
            mean and standard deviation
        X_full (ndarray):
            Full grid indices
        X_sparse (ndarray):
            Sparse grid indices
        **xi (float):
            xi constant value

    Returns:
        Acquisition function values and GP prediction (mean + stddev)
    """
    xi = kwargs.get("xi", 0.01)
    mean, sd = gpmodel.predict(X_full, verbose=0)
    mean_sample = gpmodel.predict(X_sparse, verbose=0)

    mean_sample_opt = np.nanmax(mean_sample)
    z = mean - mean_sample_opt - xi
    z = z / sd
    acq = norm.cdf(z)
    return acq, (mean, sd)



import numpy as np
import pytest
from numpy.testing import assert_
from gpim.gpreg import gpr, skgpr
from gpim import gprutils
np.random.seed(0)


def get_dummy_data():
    h = 5
    x_min, x_max = 0, 100
    y_min, y_max = 0, 100
    xx, yy = np.meshgrid(
        np.arange(x_min, x_max, h),
        np.arange(y_min, y_max, h))
    Z = np.exp(-((xx-25)**2+(yy-50)**2)/300)
    for _ in range(200):
        i = np.random.randint(Z.shape[0])
        j = np.random.randint(Z.shape[1])
        Z[i, j] = np.nan
    return Z


@pytest.mark.parametrize('kernel', ['RBF', 'Matern52'])
def test_gpr_2d(kernel):  # sanity check only, due to comput cost
    R = get_dummy_data()
    X = gprutils.get_sparse_grid(R)
    X_true = gprutils.get_full_grid(R)
    mean, sd, _ = gpr.reconstructor(
        X, R, X_true,
        kernel=kernel, learning_rate=0.1,
        iterations=2, use_gpu=False,
        verbose=False).run()
    assert_(mean.shape == sd.shape == R.shape)
    assert_(not np.isnan(mean).any())
    assert_(not np.isnan(sd).any())


@pytest.mark.parametrize('kernel', ['RBF', 'Matern52'])
def test_skgpr_2d(kernel):  # sanity check only, due to comput cost
    R = get_dummy_data()
    X = gprutils.get_sparse_grid(R)
    X_true = gprutils.get_full_grid(R)
    mean, sd, _ = skgpr.skreconstructor(
        X, R, X_true, kernel=kernel,
        learning_rate=0.1, iterations=2,
        use_gpu=False, verbose=False).run()
    assert_(mean.shape == sd.shape == R.shape)
    assert_(not np.isnan(mean).any())
    assert_(not np.isnan(sd).any())



'''
pyro_kernels.py
======
Pyro kernels
(some customized kernels TBA)
'''

import pyro.contrib.gp as gp
import pyro.distributions as dist
import torch
import warnings


def get_kernel(kernel_type, input_dim, lengthscale, use_gpu=False, **kwargs):
    """
    Initalizes one of the following kernels:
    RBF, Rational Quadratic, Matern
    Args:
        kernel_type (str):
            kernel type ('RBF', 'Rational Quadratic', Matern52')
        input_dim (int):
            number of input dimensions
            (equal to number of feature vector columns)
        lengthscale (list of two lists):
            determines lower (1st list) and upper (2nd list) bounds
            for kernel lengthscale(s). Number of elements in each list
            is equal to the input dimensions
        use_gpu (bool):
            sets default tensor type to torch.cuda.DoubleTensor
        **amplitude (list with two floats):
            determines bounds on kernel amplitude parameter
            (default is from 1e-4 to 10)
        **precision (str):
            Choose between single ('single') and double ('double') precision
    Returns:
        Pyro kernel object
    """
    precision = kwargs.get("precision", "double")
    if precision == 'single':
        tensor_type = torch.FloatTensor
        tensor_type_gpu = torch.cuda.FloatTensor
    else:
        tensor_type = torch.DoubleTensor
        tensor_type_gpu = torch.cuda.DoubleTensor

    if use_gpu and torch.cuda.is_available():
        torch.set_default_tensor_type(tensor_type_gpu)
    else:
        torch.set_default_tensor_type(tensor_type)

    amp = kwargs.get('amplitude')
    lscale = lengthscale
    amp = [1e-4, 10.] if amp is None else amp
    # Needed in Pyro < 1.0.0
    lscale_ = torch.tensor(lscale[0]) + 1e-5

    # initialize the kernel
    kernel_book = lambda input_dim: {
        'RBF': gp.kernels.RBF(
            input_dim, lengthscale=lscale_
            ),
        'RationalQuadratic': gp.kernels.RationalQuadratic(
            input_dim, lengthscale=lscale_
            ),
        'Matern52': gp.kernels.Matern52(
            input_dim, lengthscale=lscale_
            )
    }

    try:
        kernel = kernel_book(input_dim)[kernel_type]
    except KeyError:
        print('Select one of the currently available kernels:',\
              '"RBF", "RationalQuadratic", "Matern52"')
        raise

    with warnings.catch_warnings():  # TODO: use PyroSample to set priors
        warnings.filterwarnings("ignore", category=UserWarning)

        # set priors
        kernel.set_prior(
            "variance",
            dist.Uniform(
                torch.tensor(amp[0]),
                torch.tensor(amp[1])
            )
        )
        kernel.set_prior(
            "lengthscale",
            dist.Uniform(
                torch.tensor(lscale[0]),
                torch.tensor(lscale[1])
            ).independent()
        )

    return kernel



'''
gpytorch_kernels.py
======
Gpytorch kernels
(some customized kernels TBA)
'''

import gpytorch
import torch


def get_kernel(kernel_type, input_dim, on_gpu=True, **kwargs):
    """
    Initializes one of the following gpytorch kernels: RBF, Matern

    Args:
        kernel_type (str):
            Kernel type ('RBF', Matern52', 'Spectral)
        input_dim (int):
            Number of input dimensions
            (translates into number of kernel dimensions unless isotropic=True)
        on_gpu (bool):
            Sets default tensor type to torch.cuda.DoubleTensor
        **lengthscale (list of two lists):
            Determines lower (1st list) and upper (2nd list) bounds
            for kernel lengthscale(s);
            number of elements in each list is equal to the input dimensions
        **isotropic (bool):
            one kernel lengthscale in all dimensions
        **n_mixtures (int):
            number of mixtures for spectral mixture kernel
        **precision (str):
            Choose between single ('single') and double ('double') precision
    Returns:
        kernel object
    """
    precision = kwargs.get("precision", "double")
    if precision == 'single':
        tensor_type = torch.FloatTensor
        tensor_type_gpu = torch.cuda.FloatTensor
    else:
        tensor_type = torch.DoubleTensor
        tensor_type_gpu = torch.cuda.DoubleTensor

    if on_gpu and torch.cuda.is_available():
        torch.set_default_tensor_type(tensor_type_gpu)
    else:
        torch.set_default_tensor_type(tensor_type)

    lscale = kwargs.get('lengthscale')
    isotropic = kwargs.get("isotropic")
    nmix = kwargs.get("n_mixtures")
    if nmix is None:
        nmix = 4
    if lscale is not None:
        lscale = gpytorch.constraints.Interval(torch.tensor(lscale[0]),
                                               torch.tensor(lscale[1]))
    input_dim = 1 if isotropic else input_dim

    kernel_book = lambda input_dim, lscale, **kwargs: {
        'RBF': gpytorch.kernels.RBFKernel(
            ard_num_dims=input_dim,
            lengthscale_constraint=lscale
            ),
        'Matern52': gpytorch.kernels.MaternKernel(
            ard_num_dims=input_dim,
            lengthscale_constraint=lscale
            ),
        'Spectral': gpytorch.kernels.SpectralMixtureKernel(
            ard_num_dims=input_dim,
            num_mixtures=kwargs.get("nmix")
        )
    }
    try:
        kernel = kernel_book(input_dim, lscale, nmix=nmix)[kernel_type]
    except KeyError:
        print('Select one of the currently available kernels:',\
              '"RBF", "Matern52", "Spectral"')
        raise
    return kernel



import os
import numpy as np
import pytest
from numpy.testing import assert_allclose
from gpim.gpbayes.boptim import boptimizer
from gpim import gprutils
np.random.seed(0)

test_img_ei = os.path.join(
    os.path.dirname(__file__), 'test_data/test_ei.npy')
test_img_poi = os.path.join(
    os.path.dirname(__file__), 'test_data/test_poi.npy')
test_img_cb = os.path.join(
    os.path.dirname(__file__), 'test_data/test_cb.npy')


def trial_func(idx, **kwargs):
    """
    Trial function, which takes a list of indices as input
    """
    x0 = kwargs.get("x0", 5)
    y0 = kwargs.get("y0", 10)
    fwhm = kwargs.get("fwhm", 4.5)
    Z = np.exp(-4*np.log(2) * ((idx[0]-x0)**2 + (idx[1]-y0)**2) / fwhm**2)
    return Z


def initial_seed():
    """
    Creates sparse data (initial seed)
    """
    np.random.seed(0)
    x = np.arange(0, 25, 1.)
    y = x[:, np.newaxis]
    Z = trial_func([y, x])
    idx = np.random.randint(0, Z.shape[0], size=(2, 5))
    Z_sparse = np.ones_like(Z) * np.nan
    Z_sparse[idx[0], idx[1]] = Z[idx[0], idx[1]]
    return Z_sparse


@pytest.mark.parametrize(
    "acqf, result",
    [("ei", test_img_ei),
     ("poi", test_img_poi),
     ("cb", test_img_cb)])
def test_boptim(acqf, result):
    Z_sparse = initial_seed()
    X_full = gprutils.get_full_grid(Z_sparse)
    X_sparse = gprutils.get_sparse_grid(Z_sparse)
    expected_result = np.load(result)
    boptim = boptimizer(
        X_sparse, Z_sparse, X_full,
        trial_func, acquisition_function=acqf,
        exploration_steps=20,
        use_gpu=False, verbose=1)
    boptim.run()
    assert_allclose(boptim.target_func_vals[-1], expected_result)



'''
skgpr.py
======
Gaussian process regression model with a structured kernel interpolation
or a spectral mixture kernel. Serves as a high-level wrapper
for GPyTorch's (https://gpytorch.ai) Gaussian process modules with
structred kernel interpolation and spectral mixture kernel methods.

Author: Maxim Ziatdinov (email: maxim.ziatdinov@ai4microcopy.com)
'''

import time
import numpy as np
from gpim.kernels import gpytorch_kernels
from gpim import gprutils
import torch
import gpytorch
import warnings


class skreconstructor:
    """
    GP regression model with structured kernel interpolation
    or spectral mixture kernel for 2D/3D/4D image data reconstruction

    Args:
        X (ndarray):
            Grid indices with dimension :math:`c \\times N \\times M`,
            :math:`c \\times N \\times M \\times L` or
            :math:`c \\times N \\times M \\times L \\times K`,
            where *c* is equal to the number of coordinates
            (for example, for xyz coordinates, *c* = 3)
        y (ndarray):
            Observations (data points) with dimension :math:`N \\times M`,
            :math:`N \\times M \\times L` or
            :math:`N \\times M \\times L \\times K`.
            Typically, for 2D image *N* and *M* are image height and width.
            For 3D hyperspectral data *N* and *M* are spatial dimensions
            and *L* is a "spectroscopic" dimension (e.g. voltage or wavelength).
            For 4D datasets, both *L* and *K* are "spectroscopic" dimensions.
        Xtest (ndarray):
            "Test" points (for prediction with a trained GP model)
            with dimension :math:`N \\times M`, :math:`N \\times M \\times L`
            or :math:`N \\times M \\times L \\times K`
        kernel (str):
            Kernel type ('RBF' or 'Matern52')
        lengthscale (list of int  list of two list with ins):
            Determines lower (1st list) and upper (2nd list) bounds
            for kernel lengthscales. The number of elements in each list
            is equal to the dataset dimensionality.
        ski (bool):
            Perform structured kernel interpolation GP. Set to True by default.
        iterations (int):
            Number of training steps
        learning_rate (float):
            Learning rate for model training
        use_gpu (bool):
            Uses GPU hardware accelerator when set to 'True'
        verbose (int):
            Level of verbosity (0, 1, or 2)
        seed (int):
            for reproducibility
        **grid_points_ratio (float):
            Ratio of inducing points to overall points
        **n_mixtures (int):
            number of mixtures for spectral mixture kernel
        **isotropic (bool):
            one kernel lengthscale in all dimensions
        **max_root (int):
            Maximum number of Lanczos iterations to perform
            in prediction stage
        **num_batches (int):
            Number of batches for splitting the Xtest array
            (for large datasets, you may not have enough GPU memory
            to process the entire dataset at once)
        **precision (str):
            Choose between single ('single') and double ('double') precision
    """
    def __init__(self,
                 X,
                 y,
                 Xtest=None,
                 kernel='RBF',
                 lengthscale=None,
                 ski=True,
                 learning_rate=.1,
                 iterations=50,
                 use_gpu=1,
                 verbose=1,
                 seed=0,
                 **kwargs):
        """
        Initiates reconstructor parameters
        and pre-processes training and test data arrays
        """
        self.precision = kwargs.get("precision", "double")
        if self.precision == 'single':
            self.tensor_type = torch.FloatTensor
            self.tensor_type_gpu = torch.cuda.FloatTensor
        else:
            self.tensor_type = torch.DoubleTensor
            self.tensor_type_gpu = torch.cuda.DoubleTensor
        npfloat_ = np.float32 if self.precision == "single" else np.float64
        torch.manual_seed(seed)
        if use_gpu and torch.cuda.is_available():
            torch.cuda.empty_cache()
            torch.cuda.manual_seed_all(seed)
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False
            torch.set_default_tensor_type(self.tensor_type_gpu)
        input_dim = np.ndim(y)
        if Xtest is not None:
            self.fulldims = Xtest.shape[1:]
        else:
            self.fulldims = X.shape[1:]
        self.X, self.y = gprutils.prepare_training_data(
            X, y, precision=self.precision)
        if Xtest is not None:
            self.Xtest = gprutils.prepare_test_data(
                Xtest, precision=self.precision)
        self.do_ski = ski
        if kernel == "Spectral":
            self.do_ski = False
        self.toeplitz = gpytorch.settings.use_toeplitz(True)
        maxroot = kwargs.get("maxroot", 100)
        self.maxroot = gpytorch.settings.max_root_decomposition_size(maxroot)
        if use_gpu and torch.cuda.is_available():
            self.X, self.y = self.X.cuda(), self.y.cuda()
            if self.Xtest is not None:
                self.Xtest = self.Xtest.cuda()
            self.toeplitz = gpytorch.settings.use_toeplitz(False)
        else:
            torch.set_default_tensor_type(self.tensor_type)
        self.likelihood = gpytorch.likelihoods.GaussianLikelihood()
        isotropic = kwargs.get("isotropic")
        n_mixtures = kwargs.get("n_mixtures")
        if lengthscale is None and not isotropic:
            lmean = (np.mean(y.shape) / 2).astype(npfloat_)
            lengthscale = [[0. for l in range(input_dim)],
                           [lmean for l in range(input_dim)]]
        elif lengthscale is None and isotropic:
            lengthscale = [0., (np.mean(y.shape) / 2).astype(npfloat_)]
        _kernel = gpytorch_kernels.get_kernel(
            kernel, input_dim, use_gpu, lengthscale=lengthscale,
            isotropic=isotropic, precision=self.precision,
            n_mixtures=n_mixtures)
        grid_points_ratio = kwargs.get("grid_points_ratio", 1.)
        self.model = skgprmodel(self.X, self.y,
                                _kernel, self.likelihood, input_dim,
                                grid_points_ratio, self.do_ski)
        if use_gpu:
            self.model.cuda()
        self.iterations = iterations
        self.num_batches = kwargs.get("num_batches", 1)
        self.learning_rate = learning_rate

        self.noise_all = []
        if kernel == "Spectral":
            self.scales, self.means, self.weights = [], [], []
            self.hyperparams = {
                "scales": self.scales,
                "means": self.means,
                "weights": self.weights,
                "noise": self.noise_all,
                "maxdim": max(self.fulldims)
            }
        else:
            self.lscales = []
            self.hyperparams = {
                "lengthscale": self.lscales,
                "noise": self.noise_all,
            }
        self.verbose = verbose

    def train(self, **kwargs):
        """
        Training GP regression model

        Args:
            **learning_rate (float): learning rate
            **iterations (int): number of SVI training iteratons
        """
        if kwargs.get("learning_rate") is not None:
            self.learning_rate = kwargs.get("learning_rate")
        if kwargs.get("iterations") is not None:
            self.iterations = kwargs.get("iterations")
        if kwargs.get("verbose") is not None:
            self.verbose = kwargs.get("verbose")
        self.model.train()
        self.likelihood.train()
        optimizer = torch.optim.Adam(
            [{'params': self.model.parameters()}], lr=self.learning_rate)
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(
            self.likelihood, self.model)
        if self.verbose:
            print('Model training...')
        start_time = time.time()
        for i in range(self.iterations):
            optimizer.zero_grad()
            output = self.model(self.X)
            loss = -mll(output, self.y)
            loss.backward()
            optimizer.step()
            if self.do_ski:
                self.lscales.append(
                    self.model.covar_module.base_kernel.base_kernel.lengthscale.tolist()[0]
                )
            else:
                if not hasattr(self.model.covar_module, "num_mixtures"):
                    self.lscales.append(
                        self.model.covar_module.base_kernel.lengthscale.tolist()[0]
                    )
                else:
                    weights = self.model.covar_module.mixture_weights.detach()
                    scales = 1 / torch.sqrt(
                        self.model.covar_module.mixture_scales.detach())
                    means = 1 / self.model.covar_module.mixture_means.detach()
                    self.weights.append(weights.cpu().numpy().copy())
                    self.scales.append(scales.cpu().numpy().copy())
                    self.means.append(means.cpu().numpy().copy())
            self.noise_all.append(
                self.model.likelihood.noise_covar.noise.item())
            if self.verbose == 2 and (i % 10 == 0 or i == self.iterations - 1):
                if not hasattr(self.model.covar_module, "num_mixtures"):
                    template = 'iter: {} ... loss: {} ... length: {} ... noise: {} ...'
                    print(template.format(
                        i, np.around(loss.item(), 4),
                        np.around(self.lscales[-1], 4),
                        np.around(self.noise_all[-1], 7)))
                else:
                    template = 'iter: {} ... loss: {} ... noise: {} ...'
                    print(template.format(
                        i, np.around(loss.item(), 4),
                        np.around(self.noise_all[-1], 7)))
                    sort_idx = torch.argsort(weights, descending=True)
                    weights_sorted = weights[sort_idx].double()
                    scales_sorted = scales[sort_idx]
                    means_sorted = means[sort_idx]
                    print("weight" + 10*" " +  "mean" + 10*" " + "scale")
                    for w, m, s in zip(weights_sorted, means_sorted, scales_sorted):
                        print("{}  {}  {}".format(
                            w.cpu().numpy().round(4),
                            m[0].cpu().numpy().round(4),
                            s[0].cpu().numpy().round(4)))
            if self.verbose and i == 10:
                print('average time per iteration: {} s'.format(
                    np.round(time.time() - start_time, 2) / 10))
        if self.verbose:
            print('training completed in {} s'.format(
                np.round(time.time() - start_time, 2)))
            if not hasattr(self.model.covar_module, "num_mixtures"):
                print('Final parameter values:\n',
                      'lengthscale: {}, noise: {}'.format(
                        np.around(self.lscales[-1], 4),
                        np.around(self.noise_all[-1], 7)))
            else:
                print('Final parameter values:\n',
                      'noise: {}'.format(np.around(self.noise_all[-1], 7)))
                for w, m, s in zip(self.weights[-1], self.means[-1], self.scales[-1]):
                    print("{}  {}  {}".format(
                        np.around(w.astype(np.float64), 5),
                        np.around(m[0], 5),
                        np.around(s[0], 5)))
        return

    def predict(self, Xtest=None, **kwargs):
        """
        Makes a prediction with trained GP regression model

        Args:
            Xtest (ndarray):
                "Test" points (for prediction with a trained GP model)
                with dimension :math:`N \\times M`, :math:`N \\times M \\times L`
                or :math:`N \\times M \\times L \\times K`
            max_root (int):
                Maximum number of Lanczos iterations to perform
                in prediction stage
            num_batches (int):
                Number of batches for splitting the Xtest array
                (for large datasets, you may not have enough GPU memory
                to process the entire dataset at once)
        """

        def predict_(Xtest_i):
            with torch.no_grad(), gpytorch.settings.fast_pred_var(), self.toeplitz, self.maxroot:
                covar_i = self.likelihood(self.model(Xtest_i))
                mean_i, sd_i = covar_i.mean, covar_i.stddev
            return mean_i, sd_i

        if Xtest is None and self.Xtest is None:
            warnings.warn(
                "No test data provided. Using training data for prediction",
                UserWarning)
            self.Xtest = self.X
        elif Xtest is not None:
            self.Xtest = gprutils.prepare_test_data(
                Xtest, precision=self.precision)
            self.fulldims = Xtest.shape[1:]
            if next(self.model.parameters()).is_cuda:
                self.Xtest = self.Xtest.cuda()
        if kwargs.get("verbose") is not None:
            self.verbose = kwargs.get("verbose")
        if kwargs.get("num_batches") is not None:
            self.num_batches = kwargs.get("num_batches")
        if kwargs.get("max_root") is not None:
            self.max_root = kwargs.get("max_root")
        self.model.eval()
        self.likelihood.eval()
        batch_size = len(self.Xtest) // self.num_batches
        dtype_ = np.float32 if self.precision == 'single' else np.float64
        mean = np.zeros((self.Xtest.shape[0]), dtype_)
        sd = np.zeros((self.Xtest.shape[0]), dtype_)
        if self.verbose:
            print('Calculating predictive mean and uncertainty...')
        for i in range(self.num_batches):
            if self.verbose:
                print("\rBatch {}/{}".format(i+1, self.num_batches), end="")
            Xtest_i = self.Xtest[i*batch_size:(i+1)*batch_size]
            mean_i, sd_i = predict_(Xtest_i)
            mean[i*batch_size:(i+1)*batch_size] = mean_i.cpu().numpy()
            sd[i*batch_size:(i+1)*batch_size] = sd_i.cpu().numpy()
        Xtest_i = self.Xtest[(i+1)*batch_size:]
        if len(Xtest_i) > 0:
            mean_i, sd_i = predict_(Xtest_i)
            mean[(i+1)*batch_size:] = mean_i.cpu().numpy()
            sd[(i+1)*batch_size:] = sd_i.cpu().numpy()
        sd = sd.reshape(self.fulldims)
        mean = mean.reshape(self.fulldims)
        if self.verbose:
            print("\nDone")
        return mean, sd

    def run(self):
        """
        Combines train and step methods
        """
        self.train()
        mean, sd = self.predict()
        if next(self.model.parameters()).is_cuda:
            self.model.cpu()
            torch.set_default_tensor_type(self.tensor_type)
            self.X, self.y = self.X.cpu(), self.y.cpu()
            self.Xtest = self.Xtest.cpu()
            torch.cuda.empty_cache()
        return mean, sd, self.hyperparams

    def step(self, acquisition_function=None,
             batch_size=100, batch_update=False,
             lscale=None, **kwargs):
        """
        Performs single train-predict step and computes next query point with
        maximum value of acquisition function. Notice that it doesn't seem to
        work properly with a structred kernel.

        Args:
            acquisition_function (python function):
                Function that takes two parameters, mean and sd,
                and applies some math operation to them
                (e.g. :math:`\\upmu - 2 \\times \\upsigma`)
            batch_size (int):
                Number of query points to return
            batch_update:
                Filters the query points based on the specified lengthscale
            lscale (float):
                Lengthscale determining the separation (euclidean)
                distance between query points. Defaults to the kernel
                lengthscale
            **learning_rate (float):
                Learning rate for GP regression model training
            **steps (int):
                Number of SVI training iteratons
        Returns:
            Lists of indices and values for points with maximum uncertainty,
            predictive mean and standard deviation (as flattened numpy arrays)

        """
        if self.do_ski or hasattr(self.model.covar_module, "num_components"):
            raise NotImplementedError(
        "The Bayesian optimization routines are not available for structured or spectral kernel")
        if kwargs.get("learning_rate") is not None:
            self.learning_rate = kwargs.get("learning_rate")
        if kwargs.get("iterations") is not None:
            self.iterations = kwargs.get("iterations")
        if lscale is None:
            lscale = self.model.covar_module.base_kernel.lengthscale.mean().item()
        # train a model
        self.train(learning_rate=self.learning_rate, iterations=self.iterations)
        # make prediction
        mean, sd = self.predict()
        # find point with maximum value of acquisition function
        sd_ = sd.reshape(self.fulldims)
        mean_ = mean.reshape(self.fulldims)
        vals, inds = gprutils.acquisition(
            mean_, sd_, acquisition_function,
            batch_size, batch_update, lscale)
        return vals, inds, mean, sd


class skgprmodel(gpytorch.models.ExactGP):
    """
    GP regression model with structured kernel interpolation
    or spectral mixture kernel.

    Args:
        X (ndarray):
            Grid indices with dimension :math:`n \\times c`,
            where *n* is the number of observation points
            and *c* is equal to the number of coordinates
            (for example, for xyz coordinates, *c* = 3)
        y (ndarray):
            Observations (data points) with dimension n
        kernel (gpytorch kernel object):
            Kernel
        likelihood (gpytorch likelihood object):
            The Gaussian likelihood
        input_dim (int):
            Number of input dimensions
            (equal to number of feature vector columns)
        grid_points_ratio (float):
            Ratio of inducing points to overall points
    """

    def __init__(self, X, y, kernel, likelihood,
                 input_dim=3, grid_points_ratio=1.,
                 do_ski=False):
        """
        Initializes model parameters
        """
        super(skgprmodel, self).__init__(X, y, likelihood)
        self.mean_module = gpytorch.means.ConstantMean()
        if hasattr(kernel, "num_mixtures"):
            self.covar_module = kernel
            self.covar_module.initialize_from_data(X, y)
        else:
            self.covar_module = gpytorch.kernels.ScaleKernel(kernel)
        if not hasattr(kernel, "num_mixtures") and do_ski:
            grid_size = gpytorch.utils.grid.choose_grid_size(
                X, ratio=grid_points_ratio)
            self.covar_module = gpytorch.kernels.GridInterpolationKernel(
                self.covar_module, grid_size=grid_size, num_dims=input_dim)

    def forward(self, x):
        """
        Forward path
        """
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)



'''
gpr.py
======
Gaussian process regression:
model training, prediction and uncertainty exploration
This module serves as a high-level wrapper for sparse Gaussian processes module
from Pyro probabilistic programming library (https://pyro.ai/)
for easy work with scientific image (2D) and hyperspectral (3D) data.
Author: Maxim Ziatdinov (email: maxim.ziatdinov@ai4microcopy.com)
'''

import time
import numpy as np
from gpim.kernels import pyro_kernels
from gpim import gprutils
import torch
import pyro
import pyro.contrib.gp as gp
import warnings


class reconstructor:
    """
    Class for Gaussian process-based reconstuction
    of sparse 2D images and 3D spectroscopic datasets,
    and exploration/explotaition routines for the selection
    of next query point.

    Args:
        X (ndarray):
            Grid indices with dimensions :math:`c \\times N \\times M`
            or :math:`c \\times N \\times M \\times L`
            where *c* is equal to the number of coordinates
            (for example, for *xyz* coordinates, *c* = 3)
        y (ndarray):
            Observations (data points) with dimensions
            :math:`N \\times M` or :math:`N \\times M \\times L`.
            Typically, for 2D image *N* and *M* are image height and width,
            whereas for 3D hyperspectral data *N* and *M* are spatial dimensions
            and *L* is a spectorcopic dimension (e.g. voltage or wavelength).
        Xtest (ndarray):
            "Test" points (for prediction with a trained GP model)
            with dimensions :math:`N \\times M` or :math:`N \\times M \\times L`
        kernel (str):
            Kernel type ('RBF', 'Matern52', 'RationalQuadratic')
        lengthscale (list of int or list of two lists with int):
            Determines lower (1st value or 1st list) and upper (2nd value or 2nd list)
            bounds for kernel lengthscales. For list with two integers,
            the kernel will have only one lenghtscale, even if the dataset
            is multi-dimensional. For lists of two lists, the number of elements
            in each list must be equal to the dataset dimensionality.
        sparse (bool):
            Perform sparse GP regression when set to True.
        indpoints (int):
            Number of inducing points for SparseGPRegression.
            Defaults to total_number_of_points // 10.
        learning_rate (float):
            Learning rate for model training
        iterations (int): Number of SVI training iteratons
        use_gpu (bool):
            Uses GPU hardware accelerator when set to 'True'.
            Notice that for large datasets training model without GPU
            is extremely slow.
        verbose (int):
            Level of verbosity (0, 1, or 2)
        seed (int):
            for reproducibility
        **amplitude (float): kernel variance or amplitude squared
        **precision (str):
            Choose between single ('single') and double ('double') precision
        **jitter (float):
            Float between 1e-4 and 1e-6 for numerical stability
    """
    def __init__(self,
                 X,
                 y,
                 Xtest=None,
                 kernel='RBF',
                 lengthscale=None,
                 sparse=False,
                 indpoints=None,
                 learning_rate=5e-2,
                 iterations=1000,
                 use_gpu=False,
                 verbose=1,
                 seed=0,
                 **kwargs):
        """
        Initiates reconstructor parameters
        and pre-processes training and test data arrays
        """
        self.precision = kwargs.get("precision", "double")
        if self.precision == 'single':
            self.tensor_type = torch.FloatTensor
            self.tensor_type_gpu = torch.cuda.FloatTensor
        else:
            self.tensor_type = torch.DoubleTensor
            self.tensor_type_gpu = torch.cuda.DoubleTensor
        npfloat_ = np.float32 if self.precision == "single" else np.float64
        self.verbose = verbose
        torch.manual_seed(seed)
        pyro.set_rng_seed(seed)
        pyro.clear_param_store()
        if use_gpu and torch.cuda.is_available():
            torch.cuda.empty_cache()
            torch.cuda.manual_seed_all(seed)
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False
            torch.set_default_tensor_type(self.tensor_type_gpu)
            use_gpu = True
        else:
            torch.set_default_tensor_type(self.tensor_type)
            use_gpu = False
        input_dim = np.ndim(y)
        self.X, self.y = gprutils.prepare_training_data(
            X, y, precision=self.precision)
        self.do_sparse = sparse
        if lengthscale is None and not kwargs.get("isotropic"):
            lmean = (np.mean(y.shape) / 2).astype(npfloat_)
            lengthscale = [[0. for l in range(input_dim)],
                           [lmean for l in range(input_dim)]]  # TODO Make separate lscale for each dim
        elif lengthscale is None and kwargs.get("isotropic"):
            lengthscale = [0., (np.mean(y.shape) / 2).astype(npfloat_)]
        kernel = pyro_kernels.get_kernel(
            kernel, input_dim, lengthscale, use_gpu,
            amplitude=kwargs.get('amplitude'), precision=self.precision)
        if Xtest is not None:
            self.fulldims = Xtest.shape[1:]
        else:
            self.fulldims = X.shape[1:]
        if Xtest is not None:
            self.Xtest = gprutils.prepare_test_data(
                Xtest, precision=self.precision)
        else:
            self.Xtest = Xtest
        if use_gpu:
            self.X = self.X.cuda()
            self.y = self.y.cuda()
            if self.Xtest is not None:
                self.Xtest = self.Xtest.cuda()
        jitter = kwargs.get("jitter", 1.0e-5)
        if not self.do_sparse:
            self.model = gp.models.GPRegression(
                self.X,  self.y, kernel, jitter=jitter)
        else:
            if indpoints is None:
                indpoints = len(self.X) // 10
                indpoints = indpoints + 1 if indpoints == 0 else indpoints
            else:
                indpoints = len(self.X) if indpoints > len(self.X) else indpoints
            Xu = self.X[::len(self.X) // indpoints]
            if self.verbose == 2:
                print("# of inducing points for sparse GP regression: {}".format(len(Xu)))
            self.model = gp.models.SparseGPRegression(
                self.X, self.y, kernel, Xu, jitter=jitter)
        if use_gpu:
            self.model.cuda()
        self.learning_rate = learning_rate
        self.iterations = iterations
        self.hyperparams = {}
        self.indpoints_all = []
        self.lscales, self.noise_all, self.amp_all = [], [], []
        self.hyperparams = {
            "lengthscale": self.lscales,
            "noise": self.noise_all,
            "variance": self.amp_all,
            "inducing_points": self.indpoints_all
            }

    def train(self, **kwargs):
        """
        Training sparse GP regression model

        Args:
            **learning_rate (float): learning rate
            **iterations (int): number of SVI training iteratons
        """
        if kwargs.get("learning_rate") is not None:
            self.learning_rate = kwargs.get("learning_rate")
        if kwargs.get("iterations") is not None:
            self.iterations = kwargs.get("iterations")
        if kwargs.get("verbose") is not None:
            self.verbose = kwargs.get("verbose")
        pyro.clear_param_store()
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.learning_rate)
        loss_fn = pyro.infer.Trace_ELBO().differentiable_loss
        start_time = time.time()
        if self.verbose:
            print('Model training...')
        for i in range(self.iterations):
            optimizer.zero_grad()
            loss = loss_fn(self.model.model, self.model.guide)
            loss.backward()
            optimizer.step()
            self.lscales.append(self.model.kernel.lengthscale_map.tolist())
            self.amp_all.append(self.model.kernel.variance_map.item())
            self.noise_all.append(self.model.noise.item())
            if self.do_sparse:
                self.indpoints_all.append(self.model.Xu.detach().cpu().numpy())
            if self.verbose == 2 and (i % 100 == 0 or i == self.iterations - 1):
                print('iter: {} ...'.format(i),
                      'loss: {} ...'.format(np.around(loss.item(), 4)),
                      'amp: {} ...'.format(np.around(self.amp_all[-1], 4)),
                      'length: {} ...'.format(np.around(self.lscales[-1], 4)),
                      'noise: {} ...'.format(np.around(self.noise_all[-1], 7)))
            if i == 100 and self.verbose:
                print('average time per iteration: {} s'.format(
                    np.round(time.time() - start_time, 2) / 100))
        if self.verbose:
            print('training completed in {} s'.format(
                np.round(time.time() - start_time, 2)))
            print('Final parameter values:\n',
                'amp: {}, lengthscale: {}, noise: {}'.format(
                    np.around(self.model.kernel.variance_map.item(), 4),
                    np.around(self.model.kernel.lengthscale_map.tolist(), 4),
                    np.around(self.model.noise.item(), 7)))
        return

    def predict(self, Xtest=None, **kwargs):
        """
        Uses trained GP regression model to make predictions

        Args:
            Xtest (ndarray):
                "Test" points (for prediction with a trained GP model)
                with dimensions :math:`N \\times M` or :math:`N \\times M \\times L`.
                Uses Xtest from __init__ by default. If Xtest is None,
                uses training data X.
        Returns:
            Predictive mean and standard deviation
        """
        if Xtest is None and self.Xtest is None:
            warnings.warn(
                "No test data provided. Using training data for prediction",
                UserWarning)
            self.Xtest = self.X
        elif Xtest is not None:
            self.Xtest = gprutils.prepare_test_data(
                Xtest, precision=self.precision)
            self.fulldims = Xtest.shape[1:]
            if next(self.model.parameters()).is_cuda:
                self.Xtest = self.Xtest.cuda()
        if kwargs.get("verbose") is not None:
            self.verbose = kwargs.get("verbose")
        if self.verbose:
            print("Calculating predictive mean and variance...", end=" ")
        with torch.no_grad():
            mean, cov = self.model(self.Xtest, full_cov=False, noiseless=False)
        mean = mean.cpu().numpy()
        sd = cov.sqrt().cpu().numpy()
        mean = mean.reshape(self.fulldims)
        sd = sd.reshape(self.fulldims)
        if self.verbose:
            print("Done")
        return mean, sd

    def run(self, **kwargs):
        """
        Trains the initialized model and calculate predictive mean and variance

        Args:
            **learning_rate (float):
                learning rate for GP regression model training
            **steps (int):
                number of SVI training iteratons
        Returns:
            Predictive mean, standard deviation and
            dictionary with hyperparameters evolution
            as a function of SVI steps
        """
        if kwargs.get("learning_rate") is not None:
            self.learning_rate = kwargs.get("learning_rate")
        if kwargs.get("iterations") is not None:
            self.iterations = kwargs.get("iterations")
        self.train(learning_rate=self.learning_rate, iterations=self.iterations)
        mean, sd = self.predict()
        if next(self.model.parameters()).is_cuda:
            self.model.cpu()
            torch.set_default_tensor_type(self.tensor_type)
            self.X, self.y = self.X.cpu(), self.y.cpu()
            self.Xtest = self.Xtest.cpu()
            torch.cuda.empty_cache()
        return mean, sd, self.hyperparams

    def step(self, acquisition_function=None,
             batch_size=100, batch_update=False,
             lscale=None, **kwargs):
        """
        Performs single train-predict step for exploration analysis
        returning a new point with maximum value of acquisition function

        Args:
            acquisition_function (python function):
                Function that takes two parameters, mean and sd,
                and applies some math operation to them
                (e.g. :math:`\\upmu - 2 \\times \\upsigma`)
            batch_size (int):
                Number of query points to return
            batch_update:
                Filters the query points based on the specified lengthscale
            lscale (float):
                Lengthscale determining the separation (euclidean)
                distance between query points. Defaults to the kernel
                lengthscale
            **learning_rate (float):
                Learning rate for GP regression model training
            **steps (int):
                Number of SVI training iteratons
        Returns:
            Lists of indices and values for points with maximum uncertainty,
            predictive mean and standard deviation (as flattened numpy arrays)
        """
        if kwargs.get("learning_rate") is not None:
            self.learning_rate = kwargs.get("learning_rate")
        if kwargs.get("iterations") is not None:
            self.iterations = kwargs.get("iterations")
        if lscale is None:
            lscale = self.model.kernel.lengthscale.mean().item()
        # train a model
        self.train(learning_rate=self.learning_rate, iterations=self.iterations)
        # make prediction
        mean, sd = self.predict()
        # find point with maximum value of acquisition function
        sd_ = sd.reshape(self.fulldims)
        mean_ = mean.reshape(self.fulldims)
        vals, inds = gprutils.acquisition(
            mean_, sd_, acquisition_function,
            batch_size, batch_update, lscale)
        return vals, inds, mean, sd



__author__ = "Maxim Ziatdinov"
__copyright__ = "Copyright Maxim Ziatdinov (2020)"
__version__ = "0.3.6"
__maintainer__ = "Maxim Ziatdinov"
__email__ = "maxim.ziatdinov@ai4microcopy.com"
__date__ = "05/20/2020"

from setuptools import setup, find_packages
import os

module_dir = os.path.dirname(os.path.abspath(__file__))

if __name__ == "__main__":
    setup(
        name='gpim',
        version='0.3.6',
        description='Gaussian processes for image analysis',
        long_description=open(os.path.join(module_dir, 'README.md')).read(),
        long_description_content_type='text/markdown',
        url='https://github.com/ziatdinovmax/GPim',
        author='Maxim Ziatdinov',
        author_email='maxim.ziatdinov@ai4microcopy.com',
        license='MIT license',
        packages=find_packages(include=['gpim', 'gpim.*']),
        zip_safe=False,
        install_requires=[
            'numpy>=1.16.5',
            'matplotlib>=3.1.1',
            'torch>=1.3.1',
            'pyro-ppl>=0.4.1',
            'gpytorch>=0.3.6',
            'scikit-image==0.16.2'
        ],
        classifiers=['Programming Language :: Python',
                     'Development Status :: 3 - Alpha',
                     'Intended Audience :: Science/Research',
                     'Operating System :: OS Independent',
                     'Topic :: Scientific/Engineering']
    )



'''
vgpr.py
======
Gaussian process regression model for vector-valued functions.
Serves as a high-level wrapper for GPyTorch's (https://gpytorch.ai)
Gaussian processes with correlated and independent output dimensions.
Author: Maxim Ziatdinov (email: maxim.ziatdinov@ai4microcopy.com)
'''

import time
import numpy as np
from gpim import gprutils
from gpim.kernels import gpytorch_kernels
import torch
import gpytorch
import warnings


class vreconstructor:
    """
    Multi-output GP regression model for vector-valued
    2D/3D/4D functions.

    Args:
        X (ndarray):
            Grid indices with dimension :math:`c \\times N \\times M`,
            :math:`c \\times N \\times M \\times L` or
            :math:`c \\times N \\times M \\times L \\times K`,
            where *c* is equal to the number of coordinates
            (for example, for xyz coordinates, *c* = 3)
        y (ndarray):
            Observations (data points) with dimension :math:`N \\times M`,
            :math:`N \\times M \\times L \\times d` or
            :math:`N \\times M \\times L \\times K \\times d`,
            where d is a number of output dimensions.
            Typically, for 2D image *N* and *M* are image height and width.
            For 3D hyperspectral data *N* and *M* are spatial dimensions
            and *L* is a "spectroscopic" dimension (e.g. voltage or wavelength).
            For 4D datasets, both *L* and *K* are "spectroscopic" dimensions.
        Xtest (ndarray):
            "Test" points (for prediction with a trained GP model)
            with dimension :math:`N \\times M`, :math:`N \\times M \\times L`
            or :math:`N \\times M \\times L \\times K`
        kernel (str):
            Kernel type ('RBF' or 'Matern52')
        lengthscale (list of int  list of two list with ins):
            Determines lower (1st list) and upper (2nd list) bounds
            for kernel lengthscales. The number of elements in each list
            is equal to the dataset dimensionality.
        independent (bool):
            Indicates whether output dimensions are independent or correlated
        iterations (int):
            Number of training steps
        learning_rate (float):
            Learning rate for model training
        use_gpu (bool):
            Uses GPU hardware accelerator when set to 'True'
        verbose (int):
            Level of verbosity (0, 1, or 2)
        seed (int):
            for reproducibility
        **isotropic (bool):
            one kernel lengthscale in all dimensions
        **max_root (int):
            Maximum number of Lanczos iterations to perform
            in prediction stage
        **num_batches (int):
            Number of batches for splitting the Xtest array
            (for large datasets, you may not have enough GPU memory
            to process the entire dataset at once)
    """
    def __init__(self,
                 X,
                 y,
                 Xtest=None,
                 kernel='RBF',
                 lengthscale=None,
                 independent=False,
                 learning_rate=.1,
                 iterations=50,
                 use_gpu=1,
                 verbose=1,
                 seed=0,
                 **kwargs):
        """
        Initiates reconstructor parameters
        and pre-processes training and test data arrays
        """
        self.precision = kwargs.get("precision", "double")
        if self.precision == 'single':
            self.tensor_type = torch.FloatTensor
            self.tensor_type_gpu = torch.cuda.FloatTensor
        else:
            self.tensor_type = torch.DoubleTensor
            self.tensor_type_gpu = torch.cuda.DoubleTensor
        torch.manual_seed(seed)
        if use_gpu and torch.cuda.is_available():
            torch.cuda.empty_cache()
            torch.cuda.manual_seed_all(seed)
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False
            torch.set_default_tensor_type(self.tensor_type_gpu)
        input_dim = np.ndim(y) - 1
        X, y = gprutils.prepare_training_data(
            X, y, vector_valued=True, precision=self.precision)
        num_tasks = y.shape[-1]
        if Xtest is not None:
            self.fulldims = Xtest.shape[1:] + (num_tasks,)
        else:
            self.fulldims = X.shape[1:] + (num_tasks,)
        if Xtest is not None:
            Xtest = gprutils.prepare_test_data(Xtest, precision=self.precision)
        self.X, self.y, self.Xtest = X, y, Xtest
        self.toeplitz = gpytorch.settings.use_toeplitz(True)
        maxroot = kwargs.get("maxroot", 100)
        self.maxroot = gpytorch.settings.max_root_decomposition_size(maxroot)
        if use_gpu and torch.cuda.is_available():
            self.X, self.y = self.X.cuda(), self.y.cuda()
            if self.Xtest is not None:
                self.Xtest = self.Xtest.cuda()
            self.toeplitz = gpytorch.settings.use_toeplitz(False)
        else:
            torch.set_default_tensor_type(self.tensor_type)
        self.likelihood = gpytorch.likelihoods.MultitaskGaussianLikelihood(num_tasks)
        isotropic = kwargs.get("isotropic")
        _kernel = gpytorch_kernels.get_kernel(
            kernel, input_dim, use_gpu, lengthscale=lengthscale,
            isotropic=isotropic, precision=self.precision)

        if not independent:
            self.model = vgprmodel(self.X, self.y,
                                    _kernel, self.likelihood, num_tasks)
        else:
            self.model = ivgprmodel(self.X, self.y,
                                    _kernel, self.likelihood, num_tasks)

        if use_gpu:
            self.model.cuda()
        self.iterations = iterations
        self.num_batches = kwargs.get("num_batches", 1)
        self.learning_rate = learning_rate
        self.independent = independent
        self.lscales, self.noise_all = [], []
        self.hyperparams = {
            "lengthscale": self.lscales,
            "noise": self.noise_all,
        }
        self.verbose = verbose

    def train(self, **kwargs):
        """
        Training GP regression model

        Args:
            **learning_rate (float): learning rate
            **iterations (int): number of SVI training iteratons
        """
        if kwargs.get("learning_rate") is not None:
            self.learning_rate = kwargs.get("learning_rate")
        if kwargs.get("iterations") is not None:
            self.iterations = kwargs.get("iterations")
        if kwargs.get("verbose") is not None:
            self.verbose = kwargs.get("verbose")
        self.model.train()
        self.likelihood.train()
        optimizer = torch.optim.Adam(
            [{'params': self.model.parameters()}], lr=self.learning_rate)
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(
            self.likelihood, self.model)
        if self.verbose:
            print('Model training...')
        start_time = time.time()
        for i in range(self.iterations):
            optimizer.zero_grad()
            output = self.model(self.X)
            loss = -mll(output, self.y)
            loss.backward()
            optimizer.step()
            if not self.independent:
                self.lscales.append(
                    self.model.covar_module.data_covar_module.lengthscale.tolist()[0])
            else:
                self.lscales.append(
                    self.model.covar_module.base_kernel.lengthscale.tolist()[0])
            self.noise_all.append(
                self.model.likelihood.noise_covar.noise.tolist())
            if self.verbose == 2 and (i % 10 == 0 or i == self.iterations - 1):
                print('iter: {} ...'.format(i),
                      'loss: {} ...'.format(np.around(loss.item(), 4)),
                      'length: {} ...'.format(np.around(self.lscales[-1], 4)),
                      'noise: {} ...'.format(np.around(self.noise_all[-1], 7)))
            if self.verbose and i == 10:
                print('average time per iteration: {} s'.format(
                    np.round(time.time() - start_time, 2) / 10))
        if self.verbose:
            print('training completed in {} s'.format(
                np.round(time.time() - start_time, 2)))
            print('Final parameter values:\n',
                'lengthscale: {}, noise: {}'.format(
                    np.around(self.lscales[-1], 4),
                    np.around(self.noise_all[-1], 7)))
        return

    def predict(self, Xtest=None, **kwargs):
        """
        Makes a prediction with trained GP regression model

        Args:
            Xtest (ndarray):
            "Test" points (for prediction with a trained GP model)
            with dimension :math:`N \\times M`, :math:`N \\times M \\times L`
            or :math:`N \\times M \\times L \\times K`
        max_root (int):
            Maximum number of Lanczos iterations to perform
            in prediction stage
        num_batches (int):
            Number of batches for splitting the Xtest array
            (for large datasets, you may not have enough GPU memory
            to process the entire dataset at once)
        """

        def predict_(Xtest_i):
            with torch.no_grad(), gpytorch.settings.fast_pred_var(), self.toeplitz, self.maxroot:
                covar_i = self.likelihood(self.model(Xtest_i))
                samples = torch.cat([covar_i.rsample() for _ in range(n_samples)])
                samples = samples.reshape(n_samples, Xtest_i.shape[0], self.y.shape[1])
                mean_i = torch.mean(samples, dim=0)
                sd_i = torch.sqrt(torch.var(samples, dim=0))
            return mean_i, sd_i

        if Xtest is None and self.Xtest is None:
            warnings.warn(
                "No test data provided. Using training data for prediction",
                UserWarning)
            self.Xtest = self.X
        elif Xtest is not None:
            self.Xtest = gprutils.prepare_test_data(
                Xtest, precision=self.precision)
            self.fulldims = Xtest.shape[1:] + (self.y.shape[-1],)
            if next(self.model.parameters()).is_cuda:
                self.Xtest = self.Xtest.cuda()
        if kwargs.get("verbose") is not None:
            self.verbose = kwargs.get("verbose")
        if kwargs.get("num_batches") is not None:
            self.num_batches = kwargs.get("num_batches")
        if kwargs.get("max_root") is not None:
            self.max_root = kwargs.get("max_root")
        n_samples = kwargs.get("n_samples", 100)
        self.model.eval()
        self.likelihood.eval()
        batch_size = len(self.Xtest) // self.num_batches
        dtype_ = np.float32 if self.precision == 'single' else np.float64 
        mean = np.zeros((self.Xtest.shape[0], self.y.shape[-1]), dtype_)
        sd = np.zeros((self.Xtest.shape[0], self.y.shape[-1]), dtype_)
        if self.verbose:
            print('Calculating predictive mean and uncertainty...')
        for i in range(self.num_batches):
            if self.verbose:
                print("\rBatch {}/{}".format(i+1, self.num_batches), end="")
            Xtest_i = self.Xtest[i*batch_size:(i+1)*batch_size]
            mean_i, sd_i = predict_(Xtest_i)
            mean[i*batch_size:(i+1)*batch_size] = mean_i.cpu().numpy()
            sd[i*batch_size:(i+1)*batch_size] = sd_i.cpu().numpy()
        Xtest_i = self.Xtest[(i+1)*batch_size:]
        if len(Xtest_i) > 0:
            mean_i, sd_i = predict_(Xtest_i)
            mean[(i+1)*batch_size:] = mean_i.cpu().numpy()
            sd[(i+1)*batch_size:] = sd_i.cpu().numpy()
        sd = sd.reshape(self.fulldims)
        mean = mean.reshape(self.fulldims)
        if self.verbose:
            print("\nDone")
        return mean, sd

    def run(self):
        """
        Combines train and step methods
        """
        self.train()
        mean, sd = self.predict()
        if next(self.model.parameters()).is_cuda:
            self.model.cpu()
            torch.set_default_tensor_type(self.tensor_type)
            self.X, self.y = self.X.cpu(), self.y.cpu()
            self.Xtest = self.Xtest.cpu()
            torch.cuda.empty_cache()
        return mean, sd, self.hyperparams


class vgprmodel(gpytorch.models.ExactGP):
    """
    GP regression model for vector-valued functions
    with correlated output dimensions

    Args:
        X (ndarray):
            Grid indices with dimension :math:`n \\times c`,
            where *n* is the number of observation points
            and *c* is equal to the number of coordinates
            (for example, for xyz coordinates, *c* = 3)
        y (ndarray):
            Observations (data points) with dimension
            :math:`n \\times d`, where d is number of
            the function components
        kernel (gpytorch kernel object):
            'RBF' or 'Matern52' kernels
        likelihood (gpytorch likelihood object):
            The Gaussian likelihood
        num_tasks (int):
            Number of tasks (equal to number of outputs)
    """
    def __init__(self, X, y, kernel, likelihood, num_tasks):
        super(vgprmodel, self).__init__(X, y, likelihood)
        self.mean_module = gpytorch.means.MultitaskMean(
            gpytorch.means.ConstantMean(), num_tasks)
        self.covar_module = gpytorch.kernels.MultitaskKernel(kernel, num_tasks)

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultitaskMultivariateNormal(mean_x, covar_x)


class ivgprmodel(gpytorch.models.ExactGP):
    """
    GP regression model for vector-valued functions
    with independent output dimensions

    Args:
        X (ndarray):
            Grid indices with dimension :math:`n \\times c`,
            where *n* is the number of observation points
            and *c* is equal to the number of coordinates
            (for example, for xyz coordinates, *c* = 3)
        y (ndarray):
            Observations (data points) with dimension
            :math:`n \\times d`, where d is number of
            the function components
        kernel (gpytorch kernel object):
            'RBF' or 'Matern52' kernels
        likelihood (gpytorch likelihood object):
            The Gaussian likelihood
        num_tasks (int):
            Number of tasks (equal to number of outputs)
    """
    def __init__(self, X, y, kernel, likelihood, num_tasks):
        super(ivgprmodel, self).__init__(X, y, likelihood)
        self.mean_module = gpytorch.means.ConstantMean(
            batch_shape=torch.Size([num_tasks]))
        kernel.batch_shape = torch.Size([num_tasks])
        self.covar_module = gpytorch.kernels.ScaleKernel(
            kernel, batch_shape=torch.Size([num_tasks]))

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultitaskMultivariateNormal.from_batch_mvn(
            gpytorch.distributions.MultivariateNormal(mean_x, covar_x))



"""
gprutils.py
===========

Utility functions for the analysis of sparse image and hyperspectral data
with Gaussian processes.

Author: Maxim Ziatdinov (email: maxim.ziatdinov@ai4microcopy.com)
"""

import copy
import os

import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import pyro
import torch
from mpl_toolkits.mplot3d import Axes3D
from torch.distributions import constraints, transform_to


def prepare_training_data(X, y=None, vector_valued=False, **kwargs):
    """
    Reshapes and converts data to torch tensors for GP analysis

    Args:
        X (ndarray):
            Grid indices with dimensions
            :math:`c \\times N \\times M \\times L`,
            where *c* is equal to the number of coordinates
            (for example, for xyz coordinates, *c* = 3)
        y (ndarray):
            Observations (data points) with dimensions N x M x L
        **precision (str):
            Choose between single ('single') and double ('double') precision

    Returns:
        Pytorch tensors with dimensions
        :math:`N \\times M \\times L \\times c`
        and :math:`N \\times M \\times L`
    """
    precision = kwargs.get("precision", "double")
    if precision == 'single':
        tensor_type = torch.FloatTensor
    else:
        tensor_type = torch.DoubleTensor
    tor = lambda n: torch.from_numpy(n).type(tensor_type)
    X = X.reshape(X.shape[0], np.product(X.shape[1:])).T
    X = tor(X[~np.isnan(X).any(axis=1)])
    if y is None:
        return X, y
    if vector_valued:
        y = y.reshape(np.product(y.shape[:-1]), y.shape[-1])
        y = tor(y[~np.isnan(y).any(axis=1)])
    else:
        y = tor(y.flatten()[~np.isnan(y.flatten())])

    return X, y


def prepare_test_data(X, **kwargs):
    """
    Reshapes and converts data to torch tensors for GP analysis

    Args:
        X (ndarray):
            Grid indices with dimensions :math:`c \\times N \\times M \\times L`
            where *c* is equal to the number of coordinates
            (for example, for xyz coordinates, *c* = 3)
        **precision (str):
            Choose between single ('single') and double ('double') precision

    Returns:
        Pytorch tensor with dimensions :math:`N \\times M \\times L \\times c`
    """
    precision = kwargs.get("precision", "double")
    if precision == 'single':
        tensor_type = torch.FloatTensor
    else:
        tensor_type = torch.DoubleTensor
    X = X.reshape(X.shape[0], np.product(X.shape[1:])).T
    X = torch.from_numpy(X).type(tensor_type)

    return X


def get_grid_indices(R, dense_x=1.):
    """
    Returns full and sparse grid indices for 2D and 3D arrays

    Args:
        R (ndarray):
            Sparse grid measurements as 2D or 3D numpy array
        dense_x (float):
            Determines grid density
            (can be increased at prediction stage)
    """
    if np.ndim(R) > 3:
        raise NotImplementedError(
            "Currently supports only 2D and 3D arrays")
    dense_x = np.float(dense_x)
    X_full = get_full_grid(R, dense_x)
    X_sparse = get_sparse_grid(R)
    return X_full, X_sparse


def get_full_grid(R, extent=None, dense_x=1.):
    """
    Creates grid indices for 2D-4D numpy arrays

    Args:
        R (ndarray):
            Grid measurements as 2D-4D numpy array
        extent (list of lists):
            Define multi-dimensional data bounds. For example, for 2D data,
            the extent parameter is [[xmin, xmax], [ymin, ymax]]
        dense_x (float):
            Determines grid density
            (can be increased at prediction stage)

    Returns:
            Grid indices as numpy array
    """
    dense_x = np.float(dense_x)
    if np.ndim(R) == 2:
        e1, e2 = R.shape
        if extent:
            dx = extent[0][1] - extent[0][0]
            dy = extent[1][1] - extent[1][0]
            dx = dense_x / (e1//dx)
            dy = dense_x / (e2//dy)
            c1, c2 = np.mgrid[
                extent[0][0]:extent[0][1]:dx, extent[1][0]:extent[1][1]:dy]
        else:
            c1, c2 = np.mgrid[:e1:dense_x, :e2:dense_x]
        X_grid = np.array([c1, c2])
    elif np.ndim(R) == 3:
        e1, e2, e3 = R.shape
        if extent:
            dx = extent[0][1] - extent[0][0]
            dy = extent[1][1] - extent[1][0]
            dz = extent[2][1] - extent[2][0]
            dx = dense_x / (e1//dx)
            dy = dense_x / (e2//dy)
            dz = dense_x / (e3//dz)
            c1, c2 = np.mgrid[
                extent[0][0]:extent[0][1]:dx, extent[1][0]:extent[1][1]:dy,
                extent[2][0]:extent[2][1]:dz]
        else:
            c1, c2, c3 = np.mgrid[:e1:dense_x, :e2:dense_x, :e3:dense_x]
        X_grid = np.array([c1, c2, c3])
    elif np.ndim(R) == 4:
        e1, e2, e3, e4 = R.shape
        if extent:
            dx = extent[0][1] - extent[0][0]
            dy = extent[1][1] - extent[1][0]
            dz = extent[2][1] - extent[2][0]
            df = extent[3][1] - extent[3][0]
            dx = dense_x / (e1//dx)
            dy = dense_x / (e2//dy)
            dz = dense_x / (e3//dz)
            df = dense_x / (e4//df)
            c1, c2 = np.mgrid[
                extent[0][0]:extent[0][1]:dx, extent[1][0]:extent[1][1]:dy,
                extent[2][0]:extent[2][1]:dz, extent[3][0]:extent[3][1]:df]
        else:
            c1, c2, c3, c4 = np.mgrid[:e1:dense_x, :e2:dense_x, :e3:dense_x, :e4:dense_x]
        X_grid = np.array([c1, c2, c3, c4])
    else:
        raise NotImplementedError("Currently works only for 2D-4D sets")
    return X_grid


def get_sparse_grid(R, extent=None):
    """
    Returns sparse grid for sparse image data

    Args:
        R (ndarray):
            Sparse grid measurements (missing values are NaNs)

    Returns:
        Sparse grid indices
    """
    if not np.isnan(R).any():
        raise NotImplementedError(
            "Missing values in sparse data must be represented as NaNs")
    X_true = get_full_grid(R, extent)
    if np.ndim(R) == 2:
        e1, e2 = R.shape
        X = X_true.copy().reshape(2, e1*e2)
        X[:, np.where(np.isnan((R.flatten())))] = np.nan
        X = X.reshape(2, e1, e2)
    elif np.ndim(R) == 3 and not np.isnan(R[..., -1]).any():
        e1, e2, e3 = R.shape
        X = X_true.copy().reshape(3, e1*e2, e3)
        indices = np.where(np.isnan((R.reshape(e1*e2, e3))))[0]
        X[:, indices] = np.nan
        X = X.reshape(3, e1, e2, e3)
    elif np.ndim(R) == 3 and np.isnan(R[..., -1]).any():
        e1, e2, e3 = R.shape
        X = X_true.copy().reshape(3, e1*e2*e3)
        indices = np.where(np.isnan((R.reshape(e1*e2*e3))))[0]
        X[:, indices] = np.nan
        X = X.reshape(3, e1, e2, e3)
    else:
        raise NotImplementedError(
            "Currently supports only 2D and 3D sets with sparsity in xy and xyz dims")
    return X


def to_constrained_interval(state_dict, lscale, amp):
    """
    Transforms kernel's unconstrained lenghscale and variance
    to their constrained domains (intervals)

    Args:
        state_dict (dict):
            Kernel's state dictionary;
            can be obtained from self.spgr.kernel.state_dict
        lscale (list):
            List of two lists with lower and upper bound(s)
            for lenghtscale prior. Number of elements in each list
            is usually equal to the number of (independent) input dimensions
        amp (list):
            List with two floats corresponding to lower and upper
            bounds for variance (square of amplitude) prior

    Returns:
        Lengthscale and variance in the constrained domain (interval)
    """
    l_ = state_dict()['lenghtscale_map_unconstrained']
    a_ = state_dict()['variance_map_unconstrained']
    l_interval = constraints.interval(
        torch.tensor(lscale[0]), torch.tensor(lscale[1]))
    a_interval = constraints.interval(
        torch.tensor(amp[0]), torch.tensor(amp[1]))
    l = transform_to(l_interval)(l_)
    a = transform_to(a_interval)(a_)
    return l, a


def corrupt_data_xy(X_true, R_true, prob=0.5, replace_w_zeros=False):
    """
    Replaces certain % of 2D or 3D image data with NaNs;
    see gprutils.corrupt_image2d and gprutils.corrupt_image3d

    Args:
        X_true (ndarray):
            Grid indices for 2D image or 3D hyperspectral data
            (3D and 4D numpy arrays, respectively)
        R_true (ndarray):
            Observations as 2D image or 3D hyperspectral data
        prob (float):
            Controls % of data to be corrupted
            (takes values between 0 and 1)
        replace_w_zeros (bool):
            Corrupts data with zeros instead of NaNs

    Returns:
        ndarays of grid indices (3D or 4D) and observations (2D or 3D)
    """
    if np.ndim(R_true) == 2:
        X, R = corrupt_image2d(X_true, R_true, prob, replace_w_zeros)
    elif np.ndim(R_true) == 3:
        X, R = corrupt_image3d(X_true, R_true, prob, replace_w_zeros)
    else:
        raise NotImplementedError("Currently supports only 2D and 3D sets")
    return X, R


def corrupt_image2d(X_true, R_true, prob, replace_w_zeros):
    """
    Replaces certain % of 2D image data with NaNs.

    Args:
        X_true (ndarray):
            3D array with grid indices for 2D image
        R_true (ndarray):
            2D image with observations
        prob (float):
            Controls % of data to be corrupted
            (takes values between 0 and 1)
        replace_w_zeros (bool):
            Corrupts data with zeros instead of NaNs


    Returns:
        3D ndarray of grid coordinates and 2D ndarray of observatons
        where the part of points is replaced with NaNs.
    """
    e1, e2 = R_true.shape
    if np.isnan(R_true).any():
        X = X_true.copy().reshape(2, e1*e2)
        X[:, np.where(np.isnan((R_true.flatten())))] = np.nan
        X = X.reshape(2, e1, e2)
        return X, R_true
    pyro.set_rng_seed(0)
    brn = pyro.distributions.Bernoulli(prob)
    indices = [i for i in range(e1*e2) if brn.sample() == 1]
    R = R_true.copy().reshape(e1*e2)
    R[indices] = np.nan
    R = R.reshape(e1, e2)
    X = X_true.copy().reshape(2, e1*e2)
    X[:, indices] = np.nan
    X = X.reshape(2, e1, e2)
    if replace_w_zeros:
        X = np.nan_to_num(X)
        R = np.nan_to_num(R)
    return X, R


def corrupt_image3d(X_true, R_true, prob, replace_w_zeros):
    """
    Replaces certain % of 3D hyperspectral data with NaNs.
    Applies differently in xy and in z dimensions.
    Specifically, for every corrupted (x, y) point
    we remove all z values associated with this point.

    Args:
        X_true (ndarray):
            4D array with grid indices for 3D hyperspectral data
        R_true (ndarray):
            3D hyperspectral data with observations
        prob (float):
            Controls % of data to be corrupted
            (takes values between 0 and 1)
        replace_w_zeros (bool):
            Corrupts data with zeros instead of NaNs


    Returns:
        4D ndarray of grid coordinates and
        3D ndarray of observatons where
        certain % of points is replaced with NaNs
        (note that for every corrupted (x, y) point
        we remove all z values associated with this point)
    """
    e1, e2, e3 = R_true.shape
    if np.isnan(R_true).any():
        X = X_true.copy().reshape(3, e1*e2, e3)
        indices = np.where(np.isnan((R_true.reshape(e1*e2, e3))))[0]
        X[:, indices] = np.nan
        X = X.reshape(3, e1, e2, e3)
        return X, R_true
    pyro.set_rng_seed(0)
    brn = pyro.distributions.Bernoulli(prob)
    indices = [i for i in range(e1*e2) if brn.sample() == 1]
    R = R_true.copy().reshape(e1*e2, e3)
    R[indices, :] = np.nan
    R = R.reshape(e1, e2, e3)
    X = X_true.copy().reshape(3, e1*e2, e3)
    X[:, indices, :] = np.nan
    X = X.reshape(3, e1, e2, e3)
    if replace_w_zeros:
        X = np.nan_to_num(X)
        R = np.nan_to_num(R)
    return X, R


def open_edge_points(R, R_true, s=6):
    """
    Opens measured curves at the edges of FOV

    Args:
        R (ndarray):
            empty/sparse data
        R_true (ndarray):
            "ground truth"
        s (int):
            step value, which determines the density of opened edge points

    Returns:
        3D ndarray with opened edge points
    """
    e1, e2 = R_true.shape[:2]
    R[0, ::s] = R_true[0, ::s]
    R[::s, 0] = R_true[::s, 0]
    R[e1-1, s:e2-s:s] = R_true[e1-1, s:e2-s:s]
    R[s::s, e2-1] = R_true[s::s, e2-1]
    return R


def plot_kernel_hyperparams(hyperparams):
    """
    Plots evolution of kernel hyperparameters
    as a function of training steps

    Args:
        hyperparams (dict):
            dictionary with kernel hyperparameters
            (see gpreg.gpr.reconstructor)
    """
    if "weights" in hyperparams.keys():
        plot_mixture_hyperparams(hyperparams)
        return
    if 'variance' in hyperparams.keys():
        _, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=(16, 4))
    else:
        _, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 4))
    l = ax1.plot(hyperparams['lengthscale'], linewidth=3)
    ax1.set_title('lengthscale')
    ax1.set_xlabel('SVI iteration')
    ax1.set_ylabel('lengthscale (px)')
    ax1.legend(l, ('dim 1', 'dim 2', 'dim 3'))
    ax2.plot(hyperparams['noise'], linewidth=3)
    ax2.set_yscale('log')
    ax2.set_title('noise')
    ax2.set_xlabel('SVI iteration')
    ax2.set_ylabel('noise (px)')
    plt.subplots_adjust(wspace=.5)
    if 'variance' in hyperparams.keys():
        ax3.plot(hyperparams['variance'], linewidth=3)
        ax3.set_yscale('log')
        ax3.set_title('variance')
        ax3.set_xlabel('SVI iteration')
        ax3.set_ylabel('variance (px)')
    plt.show()


def plot_mixture_hyperparams(hyperparams):
    """
    Plots evolution of spectral mixture kernel hyperparameters
    as a function of training iterations

    Args:
        hyperparams (dict):
            dictionary with kernel hyperparameters
            (see gpreg.skgpr.skreconstructor)
    """
    means = hyperparams["means"]
    scales = hyperparams["scales"]
    weights = hyperparams["weights"]
    noise = hyperparams["noise"]
    maxdim = hyperparams["maxdim"]

    if scales[0].shape[-1] != 2:
        raise NotImplementedError(
            "Currently supports plotting only for 2D cases"
        )

    print("Mixture (final) weights:")
    for i, w in enumerate(weights[-1]):
        print("Component {}: w = {}".format(i, w.astype(np.float64).round(5)))

    fig, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=(21, 6))
    for i, m in enumerate(means):
        label1 = "x coordinate" if i == len(means) - 1 else None
        label2 = "y coordinate" if i == len(means) - 1 else None
        ax1.scatter(np.tile(i, len(m)), m[:, 0, 0], s=18, c=np.arange(len(m)), cmap='jet', label=label1)
        ax1.scatter(np.tile(i, len(m)), m[:, 0, 1], s=18, marker='x', c=np.arange(len(m)), cmap='jet', label=label2)
    ax1.set_xlabel("Iteration", fontsize=14)
    ax1.set_ylabel("Mixture mean/period (px)", fontsize=14)
    ax1.set_title("Mixtures mean (period)", fontsize=14)
    ax1.legend()

    for i, s in enumerate(scales):
        label1 = "x coordinate" if i == len(scales) - 1 else None
        label2 = "y coordinate" if i == len(scales) - 1 else None
        ax2.scatter(np.tile(i, len(s)), s[:, 0, 0], s=18, c=np.arange(len(s)), cmap='jet', label=label1)
        ax2.scatter(np.tile(i, len(s)), s[:, 0, 1], s=18, marker='x', c=np.arange(len(s)), cmap='jet', label=label2)
    ax2.set_xlabel("Iteration", fontsize=14)
    ax2.set_ylabel("Mixture scale (px)", fontsize=14)
    ax2.set_title("Mixtures scales", fontsize=14)
    ax2.legend()

    ax3.plot(noise, linewidth=3)
    ax3.set_ylabel("noise (px)", fontsize=14)
    ax3.set_xlabel("Iteration", fontsize=14)
    ax3.set_title("noise", fontsize=14)

    ax1.set_ylim(0, maxdim)
    ax2.set_ylim(0, maxdim)

    clrbar = np.linspace(1, len(m)).reshape(-1, 1)
    ax_ = fig.add_axes([.36, -.12, .3, .8])
    img = plt.imshow(clrbar, cmap='jet')
    plt.gca().set_visible(False)
    clrbar = plt.colorbar(img, ax=ax_, orientation='horizontal')
    clrbar.set_label('Mixture component', fontsize=14, labelpad=10)
    plt.show()


def plot_raw_data(raw_data, slice_number, pos,
                  spec_window=2, norm=False, **kwargs):
    """
    Plots hyperspectral data as 2D image
    integrated over a certain range of energy/frequency
    and selected individual spectroscopic curves

    Args:
        raw_data (3D ndarray):
            hyperspectral cube (the first two dimensions are *xy* coordinates
            and the last dimension is a "spectroscopic" dimension)
        slice_number (int):
            slice from datacube to visualize
        pos (list of lists):
            list with [x, y] coordinates of points where
            single spectroscopic curves will be extracted and visualized
        spec_window (int):
            window to integrate over in frequency dimension (for 2D "slices")
        **cmap (str):
            cmap for 2D image ("slice") plot
        **z_vec (1D ndarray):
            spectroscopic measurements values (e.g. frequency, bias)
        **z_vec_label (str):
            spectroscopic measurements label (e.g. frequency, bias voltage)
        **z_vec_units (str):
            spectroscopic measurements units (e.g. Hz, V)
    """
    cmap = kwargs.get('cmap', 'magma')
    z_vec = kwargs.get('z_vec')
    z_vec_label = kwargs.get('z_vec_label')
    z_vec_units = kwargs.get('z_vec_units')
    z_vec = np.arange(raw_data.shape[-1]) if z_vec is None else z_vec
    # colors sequence
    my_colors = ['black', 'red', 'green', 'gray', 'orange', 'blue']
    # Plotting
    s = slice_number
    spw = spec_window
    _, ax = plt.subplots(1, 2, figsize=(10, 4.5))
    ax[0].imshow(np.sum(raw_data[:, :, s-spw:s+spw], axis=-1), cmap=cmap)
    for p, col in zip(pos, my_colors):
        ax[0].scatter(p[1], p[0], c=col)
        ax[1].plot(z_vec, raw_data[p[0], p[1], :], c=col)
    ax[1].axvspan(z_vec[s-spw], z_vec[s+spw], linestyle='--', alpha=.2)
    if norm:
        ax[1].set_ylim(-0.1, 1.1)
    if z_vec_label is not None and z_vec_units is not None:
        ax[1].set_xlabel(z_vec_label+', '+z_vec_units)
        ax[1].set_ylabel('Response (arb. units)')
    ax[0].set_title('Grid spectroscopy data')
    ax[1].set_title('Individual spectroscopic curves')
    plt.subplots_adjust(wspace=.3)
    plt.show()


def plot_reconstructed_data2d(R, mean, save_fig=False, **kwargs):
    """
    Plots original and GP-reconstructed data for 2D images

    Args:
        R (2D ndarray):
            Input image for GP regression
        mean (1D ndarray):
            Predictive mean, usually an output of gpr.reconstructor or
            skgpr.skreconstructor. The array is flattened
            (the actual dimensions are the same as for R)
        **cmap (str):
            cmap for 2D image plot
        **savedir (str):
            directory to save output figure
        **filepath (str):
            name of input file (to create a unique filename for plot)
        **sparsity (float)
            indicates % of data points removed (used only for figure title)
    """

    if save_fig:
        mdir = kwargs.get('savedir', 'Output')
        if not os.path.exists(mdir):
            os.makedirs(mdir)
        fpath = kwargs.get('filepath')
    sparsity = kwargs.get('sparsity')
    cmap = kwargs.get('cmap', 'nipy_spectral')
    e1, e2 = R.shape
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 6), dpi=100)
    ax1.imshow(R, cmap=cmap, origin='bottom')
    ax2.imshow(mean.reshape(e1, e2), cmap=cmap, origin='bottom')
    ax1.set_title('Input/corrupted data')
    if sparsity:
        ax2.set_title(
            'Corrupted input data\n{}% of observations removed'.format(sparsity*100))
    else:
        ax2.set_title('Input data')
    ax2.set_title('GP reconstruction')
    if save_fig:
        if fpath:
            fig.savefig(os.path.join(mdir, os.path.basename(
                os.path.splitext(fpath)[0])))
        else:
            fig.savefig(os.path.join(mdir, 'reconstruction'))
    plt.show()


def plot_reconstructed_data3d(R, mean, sd, slice_number, pos,
                              spec_window=2, save_fig=False,
                              **kwargs):
    """
    Plots original and GP-reconstructed data for 3D images

    Args:
        R (3D ndarray):
            Input image for GP regression
        mean (1D ndarray):
            Predictive mean, usually an output of gpr.reconstructor or
            skgpr.skreconstructor. The array is flattened
            (the actual dimensions are the same as for R)
        sd (1D ndarray):
            Standard deviation
            (can be flattened; actual dimensions are the same as in R)
        slice_number (int):
            slice from datacube to visualize
        pos (list of lists):
            list with [x, y] coordinates of points where
            single spectroscopic curves will be extracted and visualized
        spec_window (int):
            window to integrate over in frequency dimension (for 2D "slices")
        **cmap (str):
            colormap for 2D image ("slices") plots
        **savedir (str):
            directory to save output figure
        **sparsity (float):
            indicates % of data points removed (used only for figure title)
        **filepath (str):
            path/name of input file (to create a unique filename for plot)
        **z_vec (1D ndarray):
            spectroscopic measurements values (e.g. frequency, bias)
        **z_vec_label (str):
            spectroscopic measurements label (e.g. frequency, bias voltage)
        **z_vec_units (str):
            spectroscopic measurements units (e.g. Hz, V)
    """
    if save_fig:
        mdir = kwargs.get('savedir')
        if mdir is None:
            mdir = 'Output'
        if not os.path.exists(mdir):
            os.makedirs(mdir)
        fpath = kwargs.get('filepath')
    sparsity = kwargs.get('sparsity')
    cmap = kwargs.get('cmap', 'nipy_spectral')
    z_vec = kwargs.get('z_vec')
    z_vec_label = kwargs.get('z_vec_label')
    z_vec_units = kwargs.get('z_vec_units')
    z_vec = np.arange(R.shape[-1]) if z_vec is None else z_vec
    s = slice_number
    e1, e2, e3 = R.shape
    spw = spec_window
    Rtest = mean.reshape(e1, e2, e3)
    R_sd = sd.reshape(e1, e2, e3)
    my_colors = ['black', 'red', 'green', 'gray', 'orange', 'blue']
    fig, ax = plt.subplots(2, 2, figsize=(14, 14))
    ax[0, 0].imshow(
        np.sum(R[:, :, s-spw:s+spw], axis=-1), cmap=cmap)
    for p, col in zip(pos, my_colors):
        ax[0, 0].scatter(p[1], p[0], c=col)
        ax[0, 1].plot(z_vec, R[p[0], p[1], :], c=col)
    ax[0, 1].axvspan(z_vec[s-spw], z_vec[s+spw], linestyle='--', alpha=.15)
    ax[0, 1].set_ylim(-0.1, 1.1)
    if z_vec_label is not None and z_vec_units is not None:
        ax[0, 1].set_xlabel(z_vec_label+', '+z_vec_units)
        ax[0, 1].set_ylabel('Response (arb. units)')
    for _ax in [ax[0, 0], ax[0, 1]]:
        if sparsity:
            _ax.set_title(
                'Corrupted input data\n{}% of observations removed'.format(sparsity*100))
        else:
            _ax.set_title('Input data')
    ax[1, 0].imshow(
        np.sum(Rtest[:, :, s-spw:s+spw], axis=-1), cmap=cmap)
    for p, col in zip(pos, my_colors):
        ax[1, 0].scatter(p[1], p[0], c=col)
        ax[1, 1].plot(z_vec, Rtest[p[0], p[1], :], c=col)
        ax[1, 1].fill_between(z_vec,
                        (Rtest[p[0], p[1], :] -
                         2.0 * R_sd[p[0], p[1], :]),
                        (Rtest[p[0], p[1], :]
                         + 2.0 * R_sd[p[0], p[1], :]),
                        color=col, alpha=0.15)
    ax[1, 1].axvspan(z_vec[s-spw], z_vec[s+spw], linestyle='--', alpha=.15)
    ax[1, 1].set_ylim(-0.1, 1.1)
    if z_vec_label is not None and z_vec_units is not None:
        ax[1, 1].set_xlabel(z_vec_label+', '+z_vec_units)
        ax[1, 1].set_ylabel('Response (arb. units)')
    for _ax in [ax[1, 0], ax[1, 1]]:
        _ax.set_title('GPR reconstruction')
    plt.subplots_adjust(hspace=.3)
    if save_fig:
        if fpath:
            fig.savefig(os.path.join(mdir, os.path.basename(
                os.path.splitext(fpath)[0])))
        else:
            fig.savefig(os.path.join(mdir, 'reconstruction'))
    plt.show()


def plot_exploration_results(R_all, mean_all, sd_all, R_true,
                             episodes, slice_number, pos, dist_edge,
                             spec_window=2, mask_predictions=False,
                             **kwargs):
    """
    Plots predictions at different stages ("episodes")
    of maximum uncertainty-based sample exploration with GP

    Args:
        R_all (list with ndarrays):
            Observed data points at each exploration step
        mean_all (list of ndarrays):
            Predictive mean at each exploration step
        sd_all (list of ndarrays):
            Integrated (along energy dimension) SD at each exploration step
        R_true (ndarray):
            3D array with ground truth data (full observations) for simulated
            experiment OR a 3D array of zeros/NaNs for real experiment
        episodes (list of ints):
            list with the numbers indicating which iteration steps to visualize
        slice_number (int):
            slice from datacube to visualize
        pos (list of lists):
            list with [x, y] coordinates of points where
            single spectroscopic curves will be extracted and visualized
        dist_edge (list with two integers):
            this should be the same as in exploration analysis
        spec_win (int):
            window to integrate over in frequency dimension (for 2D "slices")
        mask_predictions (bool):
            mask edge regions not used in max uncertainty evaluation
            in predictive mean plots
        **sparsity (float):
            indicates % of data points removed (used only for figure title)
        **z_vec (1D ndarray):
            spectroscopic measurements values (e.g. frequency, bias)
        **z_vec_label (str):
            spectroscopic measurements label (e.g. frequency, bias voltage)
        **z_vec_units (str):
            spectroscopic measurements units (e.g. Hz, V)
    """

    s = slice_number
    spw = spec_window
    e1, e2, e3 = R_true.shape
    z_vec = kwargs.get('z_vec')
    z_vec_label = kwargs.get('z_vec_label')
    z_vec_units = kwargs.get('z_vec_units')
    z_vec = np.arange(e3) if z_vec is None else z_vec
    _colors = ['black', 'red', 'green', 'blue', 'orange']
    # plot ground truth data if available
    if not np.isnan(R_true).any() or np.unique(R_true).any():
        _, ax = plt.subplots(1, 2, figsize=(7, 3), dpi=100)
        ax[0].imshow(np.sum(R_true[:, :, s-spw:s+spw], axis=-1), cmap='jet')
        for p, col in zip(pos, _colors):
            ax[0].scatter(p[1], p[0], c=col)
            ax[1].plot(z_vec, R_true[p[0], p[1], :], c=col)
        ax[1].axvspan(z_vec[s-spw], z_vec[s+spw], linestyle='--', alpha=.2)
        ax[1].set_ylim(-0.1, 1.1)
        if z_vec_label is not None and z_vec_units is not None:
            ax[1].set_xlabel(z_vec_label+', '+z_vec_units)
            ax[1].set_ylabel('Response (arb. units)')
        ax[0].set_title('Grid spectroscopy\n(ground truth)')
        ax[1].set_title('Individual spectroscopic curves\n(ground truth)')

    # Plot predictions
    n = len(episodes) + 1
    fig = plt.figure(figsize=(20, 17), dpi=100)

    for i in range(1, n):
        Rcurr = R_all[episodes[i-1]].reshape(e1, e2, e3)
        Rtest = mean_all[episodes[i-1]].reshape(e1, e2, e3)
        R_sd = sd_all[episodes[i-1]].reshape(e1, e2, e3)

        ax = fig.add_subplot(4, n, i)
        ax.imshow(np.sum(Rcurr[:, :, s-spw:s+spw], axis=-1), cmap='jet')
        ax.set_title('Observations (step {})'.format(episodes[i-1]))

        ax = fig.add_subplot(4, n, i + n)
        Rtest_to_plot = copy.deepcopy((np.sum(Rtest[:, :, s-spw:s+spw], axis=-1)))
        mask = np.zeros(Rtest_to_plot.shape, bool)
        mask[dist_edge[0]:e1-dist_edge[0],
             dist_edge[1]:e2-dist_edge[1]] = True
        if mask_predictions:
            Rtest_to_plot[~mask] = np.nan
        ax.imshow(Rtest_to_plot, cmap='jet')
        for p, col in zip(pos, _colors):
            ax.scatter(p[1], p[0], c=col)
        ax.set_title('GPR reconstruction (step {})'.format(episodes[i-1]))
        ax = fig.add_subplot(4, n, i + 2*n)
        for p, col in zip(pos, _colors):
            ax.plot(z_vec, Rtest[p[0], p[1], :], c=col)
            ax.fill_between(z_vec,
                            (Rtest[p[0], p[1], :] - 2.0 *
                            R_sd[p[0], p[1], :]),
                            (Rtest[p[0], p[1], :] + 2.0 *
                            R_sd[p[0], p[1], :]),
                            color=col, alpha=0.15)
            ax.axvspan(z_vec[s-spw], z_vec[s+spw], linestyle='--', alpha=.15)
        ax.set_ylim(-0.1, 1.1)
        if z_vec_label is not None and z_vec_units is not None:
            ax.set_xlabel(z_vec_label+', '+z_vec_units)
            ax.set_ylabel('Response (arb. units)')
        ax.set_title('GPR reconstruction (step {})'.format(episodes[i-1]))

        ax = fig.add_subplot(4, n, i + 3*n)
        R_sd_to_plot = copy.deepcopy(R_sd)
        R_sd_to_plot = np.sum(R_sd_to_plot, axis=-1)
        R_sd_to_plot[~mask] = np.nan
        ax.imshow(R_sd_to_plot, cmap='jet')
        ax.set_title('Integrated uncertainty (step {})'.format(episodes[i-1]))

    plt.subplots_adjust(hspace=.4)
    plt.subplots_adjust(wspace=.3)
    plt.show()


def plot_inducing_points(hyperparams, **kwargs):
    """
    Plots inducing points evolution during training
    """
    dims_ = hyperparams['inducing_points'][0].shape[-1]
    if dims_ == 2:
        plot_inducing_points_2d(hyperparams, **kwargs)
    elif dims_ == 3:
        plot_inducing_points_3d(hyperparams, **kwargs)
    else:
        raise NotImplementedError('Supports only 2D and 3D datasets')


def plot_inducing_points_2d(hyperparams, **kwargs):
    """
    Plots 2D trajectories if inducing points

    Args:
        hyperparams (dict):
            Dictionary of hyperparameters
        **plot_from (int):
            plot from specific step
        **plot_to (int):
            plot till specific step
        **slice_step (int):
            plot every nth inducing point
    """
    learned_inducing_points = hyperparams['inducing_points']
    indp_nth = kwargs.get('slice_step')
    plot_from, plot_to = kwargs.get('plot_to'), kwargs.get('plot_from')
    if plot_from is None:
        plot_from = 0
    if plot_to is None:
        plot_to = len(learned_inducing_points)
    if indp_nth is None:
        indp_nth = 1
    fig = plt.figure(figsize=(20, 9))
    ax = fig.add_subplot(121)
    ax.set_xlabel('x coordinate (px)', fontsize=14)
    ax.set_ylabel('y coordinate (px)', fontsize=14)
    ax.set_title('Evolution of inducing points', fontsize=16)
    ax.set_aspect('auto')# 'equal' doesn't work in matplotlib 3.1.1
    colors = plt.cm.jet(
        np.linspace(0, 1,len(learned_inducing_points[plot_from:plot_to]))
    )
    for xy, c in zip(learned_inducing_points[plot_from:plot_to], colors):
        y, x = xy.T
        ax.scatter(x[::indp_nth], y[::indp_nth], c=[c], s=.15)
    clrbar = np.linspace(
        0, len(learned_inducing_points[plot_from:plot_to])).reshape(-1, 1)
    ax2 = fig.add_axes([.42, .1, .1, .8])
    img = plt.imshow(clrbar, cmap="jet")
    plt.gca().set_visible(False)
    clrbar_ = plt.colorbar(img, ax=ax2, orientation='vertical')
    clrbar_.set_label('SVI iterations', fontsize=14, labelpad=10)
    plt.show()


def plot_inducing_points_3d(hyperparams, **kwargs):
    """
    Plots 3D trajectories if inducing points during model training

    Args:
        hyperparams (dict):
            dictionary of hyperparameters
        plot_from (int):
            plot from specific step
        plot_to (int):
            plot till specific step
        slice_step (int):
            plot every nth inducing point
    """
    learned_inducing_points = hyperparams['inducing_points']
    indp_nth = kwargs.get('slice_step')
    plot_from, plot_to = kwargs.get('plot_to'), kwargs.get('plot_from')
    if plot_from is None:
        plot_from = 0
    if plot_to is None:
        plot_to = len(learned_inducing_points)
    if indp_nth is None:
        indp_nth = 1
    fig = plt.figure(figsize=(22, 9))
    ax = fig.add_subplot(121, projection='3d')
    ax.view_init(20, 30)
    ax.set_xlabel('x coordinate (px)', fontsize=14)
    ax.set_ylabel('y coordinate (px)', fontsize=14)
    ax.set_zlabel('frequency (px)', fontsize=14)
    ax.xaxis.labelpad = 10
    ax.yaxis.labelpad = 10
    ax.zaxis.labelpad = 5
    ax.set_title('Evolution of inducing points', fontsize=16)
    ax.dist = 10 # use it to zoom in/out
    ax.set_aspect('auto')# 'equal' doesn't work in matplotlib 3.1.1
    colors = plt.cm.jet(
        np.linspace(0,1,len(learned_inducing_points[plot_from:plot_to]))
    )
    for xyz, c in zip(learned_inducing_points[plot_from:plot_to], colors):
        x, y, z = xyz.T
        ax.scatter(x[::indp_nth], y[::indp_nth], z[::indp_nth], c=[c], s=.15)
    clrbar = np.linspace(
        0, len(learned_inducing_points[plot_from:plot_to])).reshape(-1, 1)
    ax2 = fig.add_axes([.37, .1, .1, .8])
    img = plt.imshow(clrbar, cmap="jet")
    plt.gca().set_visible(False)
    clrbar_ = plt.colorbar(img, ax=ax2, orientation='vertical')
    clrbar_.set_label('SVI iterations', fontsize=14, labelpad=10)
    plt.show()


def plot_query_points(inds_all, **kwargs):
    """
    Plots the exploration path (all the query points)
    in GP-based Bayesian optimization. Currently supports only 2D data.

    Args:
        inds_all (list): list of indices
        **cmap (str): colormap
    """
    cmap = kwargs.get("cmap", "cool")
    plot_lines = kwargs.get("plot_lines", False)
    inds_all = np.array(inds_all)  # transform list to ndarray for plotting
    cvals = np.arange(len(inds_all))
    clrbar = np.linspace(0, len(inds_all)).reshape(-1, 1)
    fig, ax1 = plt.subplots(1, 1, figsize=(6, 6))
    ax1.scatter(inds_all[:, 1], inds_all[:, 0], c=cvals, cmap=cmap)
    if plot_lines:
        ax1.plot(inds_all[:, 1], inds_all[:, 0])
    ax2 = fig.add_axes([.78, .1, .2, .8])
    img = plt.imshow(clrbar, cmap)
    plt.gca().set_visible(False)
    clrbar_ = plt.colorbar(img, ax=ax2, orientation='vertical')
    clrbar_.set_label('Exploration steps', fontsize=14, labelpad=10)
    plt.show()



"""
boptim.py
===========

Utility functions for the Gaussian process-based
Bayesian optimization for selecting the next query points in
images and image-like data.

Author: Maxim Ziatdinov (email: maxim.ziatdinov@ai4microcopy.com)
"""

import types
import torch
import numpy as np
from scipy import spatial
from gpim.gpreg import gpr
from gpim.gpbayes import acqfunc
from gpim import gprutils


class boptimizer:
    """
    Gaussian process-based Bayesian optimization for selecting
    next point(s) to measure/evaluate. The Bayesian optimization strategy
    consists of: i) defining prior and posterior distributions over
    the objective (target) function :math:`f` using GP; ii) using the posterior
    to derive an acquistion function :math:`\\alpha (x)`; iii) using the
    acquisition function to derive the next query point according to
    :math:`x_{next}=argmax(\\alpha (x))`; iv) evaluating :math:`f` in
    :math:`x_{next}` and updating the posterior.

    Args:
        X_seed (ndarray):
            Initial seed of (sparse) grid indices with dimensions
            :math:`c \\times N \\times M` or :math:`c \\times N \\times M \\times L`
            where *c* is equal to the number of coordinates
            (for example, for *xyz* coordinates, *c* = 3). Missing coordinates are NaNs.
            Can be potentially larger than 3D but this has not beed tested yet.
        y_seed (ndarray):
            Initial seed with sparse "observations" with dimensions
            :math:`N \\times M` or :math:`N \\times M \\times L`.
            Typically, for 2D image *N* and *M* are image height and width,
            whereas for 3D hyperspectral data *N* and *M* are spatial dimensions
            and *L* is a spectorcopic dimension (e.g. voltage or wavelength).
            Can be potentially larger than 3D but this has not beed tested yet.
        X_full (ndarray):
            Full grid indices (for prediction with a trained GP model)
            with dimensions :math:`N \\times M` or :math:`N \\times M \\times L`.
            Can be potentially larger than 3D but this has not beed tested yet.
        target_function (python function):
            Target (or objective) function. Takes a list of indices and
            returns a function value as a float.
        acquisition_function (str or python function):
            Acquisition function choise.'cb' is confidence bound, 'ei' is
            expected improvement, 'poi' is a probability of improvement.
            One can also pass a custom function, which takes GP model,
            full grid and sparse grid as parameters and returns
            acquisition function values with GP prediction (mean + sd)
        exploration_steps (int):
            Number of exploration-exploitation steps
            (the exploitation-exploration trade-off is
            determined by the acquisition function)
        batch_size (int):
                Number of query points in one batch.
                Returns a single query point when
                batch_update parameter is set to False (default)
        batch_update (bool):
            Filters the query points based on the kernel lengthscale.
            Returns a batch of points when set to True.
            The number of points in the batch may be different from
            batch_size as points are filtered based on the lengthscale.
            Defaults to False.
        kernel (str):
            Kernel type ('RBF', 'Matern52', 'RationalQuadratic').
            Defaults to 'RBF'.
        lengthscale (list of int or list of 2 lists with int):
            Determines lower (1st value or 1st list) and upper (2nd value or 2nd list)
            bounds for kernel lengthscales. For list with two integers,
            the kernel will have only one lenghtscale, even if the dataset
            is multi-dimensional. For lists of two lists, the number of elements
            in each list must be equal to the dataset dimensionality.
        sparse (bool):
            Uses inducing points-based sparse GP regression when set to True.
            False by default.
        indpoints (int):
            Number of inducing points for SparseGPRegression.
            Defaults to total_number_of_points // 10.
        learning_rate (float):
            Learning rate for GP model training
        gp_iterations (int):
            Number of SVI training iteratons for GP model
        seed (int):
            for reproducibility
        **alpha (float or int):
            alpha coefficient in the 'confidence bound' acquisition function
            (Default: 0)
        **beta (float or int):
            beta coefficient in the 'confidence bound' acquisition function
            (Default: 1)
        **xi (float):
            xi coefficient in 'expected improvement'
            and 'probability of improvement' acquisition functions
            (Default: 0.01)
        **use_gpu (bool):
            Uses GPU hardware accelerator when set to 'True'.
            Notice that for large datasets training model without GPU
            is extremely slow. At the same time, for small and (very) sparse data
            there may be no much benefit.
        **precision (str):
            "single" ot "double" floating point precision
        **jitter (float):
            Float between 1e-4 and 1e-6 for numerical stability
        **isotropic (bool):
            Use single lenghtscale for all dimensions in GP (Default: False)
        **mask (ndarray):
            Mask of ones and NaNs (NaNs are values that are not counted when
            searching for acquisition function maximum).
        **dscale (float):
            Distance parameter used in boptimizer.checkvalues or in
            boptimizer.update_points, For boptimizer.checkvalues,
            it is used in conjuction with 'gamma' and 'memory'
            parameters to select the next query point using the information
            about the location of previous points. It can be understood as a
            short-term memory, where the point at step :math:`s` is chosen no closer
            than :math:`d` to the point selected at step :math:`s-1`, :math:`\\gamma d`
            to the point selected at step :math:`s-2`, :math:`\\gamma^2 d`
            at step :math:`s-3`, and so on. Defaults to 0. For boptimizer.update_points,
            it is used to return a batch of points for a given step, which are no closer
            to each other than dscale value. Defauts to kernel average lenghtscale
            at this step.
        **gamma (float):
            gamma coefficient, value between 0 and 1.
            Used in boptimizer.checkvalues together with a 'dscale' parameter
            to determine how close the next query point can be to
            the previously selected points.
        **memory (int):
            Number of previously selected points to account for when using 'dscale'.
        **exit_strategy (0 or 1):
            Exit strategy for boptimizer.checkvalues when none
            of the points satisfies the imposed selection criteria.
            1 means that a random value will be chosen, while 0
            means that the last point in the list will be chosen
            (the length of the list is defined by 'batch_size' parameter)
        **extent(list of lists):
            Define bounds for multi-dimensional data. For example, for 2D data,
            the extent parameter is [[xmin, xmax], [ymin, ymax]]
        **save_checkpoints (bool):
            save results to disk after each iteration
        **filename (str):
            name of a file to save checkpoints to
        **verbose (int):
            Level of verbosity (0, 1, or 2)
    """
    def __init__(self,
                 X_seed,
                 y_seed,
                 X_full,
                 target_function,
                 acquisition_function='cb',
                 exploration_steps=10,
                 batch_size=100,
                 batch_update=False,
                 kernel='RBF',
                 lengthscale=None,
                 sparse=False,
                 indpoints=None,
                 gp_iterations=1000,
                 seed=0,
                 **kwargs):
        """
        Initiates Bayesian optimizer parameters
        """
        self.verbose = kwargs.get("verbose", 1)
        self.use_gpu = kwargs.get("use_gpu", False)
        learning_rate = kwargs.get("learning_rate", 5e-2)
        jitter = kwargs.get("jitter", 1.0e-6)
        isotropic = kwargs.get("isotropic", False)
        self.precision = kwargs.get("precision", "double")

        if self.use_gpu and torch.cuda.is_available():
            if self.precision == "single":
                torch.set_default_tensor_type(torch.cuda.FloatTensor)
            else:
                torch.set_default_tensor_type(torch.cuda.DoubleTensor)
        else:
            if self.precision == "single":
                torch.set_default_tensor_type(torch.FloatTensor)
            else:
                torch.set_default_tensor_type(torch.DoubleTensor)

        self.surrogate_model = gpr.reconstructor(
            X_seed, y_seed, X_full, kernel, lengthscale, sparse, indpoints,
            learning_rate, gp_iterations, self.use_gpu, self.verbose, seed,
            isotropic=isotropic, precision=self.precision, jitter=jitter)

        self.X_sparse = X_seed.copy()
        self.y_sparse = y_seed.copy()
        self.X_full = X_full

        self.target_function = target_function
        self.acquisition_function = acquisition_function
        self.exploration_steps = exploration_steps
        self.batch_update = batch_update
        self.batch_size = batch_size
        self.simulate_measurement = kwargs.get("simulate_measurement", False)
        if self.simulate_measurement:
            self.y_true = kwargs.get("y_true")
            if self.y_true is None:
                raise AssertionError(
                    "To simulate measurements, add ground truth ('y_true)")
        self.extent = kwargs.get("extent", None)
        self.alpha, self.beta = kwargs.get("alpha", 0), kwargs.get("beta", 1)
        self.xi = kwargs.get("xi", 0.01)
        self.dscale = kwargs.get("dscale", None)
        self.gamma = kwargs.get("gamma", 0.8)
        self.points_mem = kwargs.get("memory", 10)
        self.exit_strategy = kwargs.get("exit_strategy", 0)
        self.mask = kwargs.get("mask", None)
        self.save_checkpoints = kwargs.get("save_checkpoints", False)
        self.filename = kwargs.get("filename", "./boptim_results")
        self.indices_all, self.vals_all = [], []
        self.target_func_vals, self.gp_predictions = [y_seed.copy()], []

    def update_posterior(self):
        """
        Updates GP posterior
        """
        X_sparse_new, y_sparse_new = gprutils.prepare_training_data(
            self.X_sparse, self.y_sparse, precision=self.precision)
        if self.use_gpu and torch.cuda.is_available():
            torch.cuda.empty_cache()
            X_sparse_new, y_sparse_new = X_sparse_new.cuda(), y_sparse_new.cuda()
        self.surrogate_model.model.X = X_sparse_new
        self.surrogate_model.model.y = y_sparse_new
        self.surrogate_model.train(verbose=self.verbose)
        return

    def evaluate_function(self, indices, y_measured=None):
        """
        Evaluates target function in the new point(s)
        """
        indices = [indices] if not self.batch_update else indices
        if self.simulate_measurement:
            for idx in indices:
                self.y_sparse[tuple(idx)] = self.y_true[tuple(idx)]
        elif y_measured is not None:
            for idx in indices:
                self.y_sparse[tuple(idx)] = y_measured[tuple(idx)]
        else:
            for idx in indices:
                if self.extent is not None:
                    _idx = []
                    for i, e in zip(idx, self.extent):
                        _idx.append(i + e[0])
                    _idx = tuple(_idx)
                else:
                    _idx = tuple(idx)
                self.y_sparse[tuple(idx)] = self.target_function(_idx)
        self.X_sparse = gprutils.get_sparse_grid(self.y_sparse, self.extent)
        self.target_func_vals.append(self.y_sparse.copy())
        return

    def next_point(self):
        """
        Calculates next query point(s)
        """
        if self.verbose:
            print("Computing acquisition function...")
        if self.acquisition_function == 'cb':
            acq, pred = acqfunc.confidence_bound(
                self.surrogate_model, self.X_full,
                alpha=self.alpha, beta=self.beta)
        elif self.acquisition_function == 'ei':
            acq, pred = acqfunc.expected_improvement(
                self.surrogate_model, self.X_full,
                self.X_sparse, xi=self.xi)
        elif self.acquisition_function == 'poi':
            acq, pred = acqfunc.probability_of_improvement(
                self.surrogate_model, self.X_full,
                self.X_sparse, xi=self.xi)
        elif isinstance(self.acquisition_function, types.FunctionType):
            acq, pred = self.acquisition_function(
                self.surrogate_model, self.X_full, self.X_sparse)
        else:
            raise NotImplementedError(
                "Choose between 'cb', 'ei', and 'poi' acquisition functions or define your own")
        self.gp_predictions.append(pred)
        if self.mask is None:
            indices_list = np.unravel_index(np.argsort(acq.ravel()), acq.shape)
            vals_list = acq[indices_list][::-1][:self.batch_size].tolist()
            indices_list = np.dstack(indices_list)[0][::-1][:self.batch_size].tolist()
        else:
            acq = self.mask*acq
            indices_list = np.unravel_index(np.argsort(acq.ravel()), acq.shape)
            vals_list = acq[indices_list]
            vals_list = vals_list[~np.isnan(vals_list)][::-1]
            indices_list = np.dstack(indices_list)[0]
            indices_list = indices_list[:len(vals_list)][::-1]
            vals_list = vals_list[:self.batch_size].tolist()
            indices_list = indices_list[:self.batch_size].tolist()
        if not self.batch_update:
            return vals_list, indices_list
        if self.dscale is None:
            dscale_ = self.surrogate_model.model.kernel.lengthscale.mean().item()
        else:
            dscale_ = self.dscale
        vals_list, indices_list = self.update_points(
            np.array(vals_list),
            np.vstack(indices_list),
            dscale_)
        return vals_list, indices_list

    @classmethod
    def update_points(cls, acqfunc_values, indices, dscale):
        """
        Takes arrays of query points (indices and values) corresponding to
        first *n* max values of the acquisition function and returns a batch
        two lists with updated query points (indices and values) whose
        separation distance is determined by kernel lengthscale.
        Notice that the updated lists will contain less points
        than the original ones.

        Args:
            acqfunc_values (ndarray):
                (*N*,) numpy array with values of acquisition function
            indices (ndarray):
                (*N*, *c*) numpy array with corresponding indices,
                where c ia a number of dimensions of the dataset
            dscale (float):
                kernel lengthscale

        Returns:
            Tuple with computed indices and corresponding values
        """
        minval = acqfunc_values.min()
        new_max = acqfunc_values.max()
        new_max_id = np.argmax(acqfunc_values)
        max_val_all, max_id_all = [], []
        ck = indices[new_max_id]
        tree = spatial.cKDTree(indices)
        while new_max > minval - 1:
            max_val_all.append(new_max)
            max_id_all.append(new_max_id)
            nn_indices = tree.query_ball_point(ck, dscale)
            acqfunc_values[nn_indices] = minval - 1
            new_max = acqfunc_values.max()
            new_max_id = np.argmax(acqfunc_values)
            ck = indices[new_max_id]
        return max_val_all, indices[max_id_all].tolist()

    def checkvalues(self, idx_list, val_list):
        """
        Checks if a current point was already used or if the euclidian
        distances from the current point to the previous points
        are above a value controlled by parameter 'alpha',
        and if so then it selects the next-in-line point.

        Args:
            idx_list (list of lists with ints):
                Indices for max values of acquisition function;
                the list is ordered (max -> min)
            val_list (list with floats):
                Standard deviation values for each index in idx_list

        Returns:
            The first element in the input list (idx_list)
            if no previous occurences found.
            Otherwise, returns the next/closest value from the list.
        """

        def dist(idx):
            idx_prev = self.indices_all[-self.points_mem:]
            # Calculate distances between current point and previous n points
            d_all = [np.linalg.norm(np.array(idx) - np.array(i)) for i in idx_prev]
            # Calculate weighting coefficient for each distance
            dscale_all = [dscale_*self.gamma**i for i in range(len(idx_prev))]
            # Check if each distance satisfies the imposed criteria
            bool_ = 0 in [d > l for (d, l) in zip(d_all[::-1], dscale_all)]
            return bool_

        dscale_ = 0 if self.dscale is None else self.dscale
        _idx = 0
        if self.verbose == 2:
            print('Acquisition function max value {} at {}'.format(
                val_list[_idx], idx_list[_idx]))
        if len(self.indices_all) == 0:
            return idx_list[_idx], val_list[_idx]
        while (1 in [1 for a in self.indices_all if a == idx_list[_idx]]
                or dist(idx_list[_idx])):
            if self.verbose == 2:
                print("Finding the next max point...")
            _idx = _idx + 1
            if _idx == len(idx_list):
                _idx = np.random.randint(0, len(idx_list)) if self.exit_strategy else -1
                if self.verbose == 2:
                    print('Index out of list. Exiting with acquisition function value {} at {}'.format(
                        val_list[_idx], idx_list[_idx]))
                break
            if self.verbose == 2:
                print('Acquisition function max value {} at {}'.format(
                    val_list[_idx], idx_list[_idx]))
        return idx_list[_idx], val_list[_idx]

    def single_step(self, *args):
        """
        Single Bayesian optimization step
        """
        e = args[0]
        if self.verbose:
            print("\nExploration step {} / {}".format(
                e+1, self.exploration_steps))
        # train with seeded data
        if e == 0:
            self.surrogate_model.train()
        # calculate acquisition function and get next query points
        vals, inds = self.next_point()
        if not self.batch_update:
            inds, vals = self.checkvalues(inds, vals)
        # evaluate function
        self.evaluate_function(inds)
        # update posterior
        self.update_posterior()
        # store indices and values
        self.indices_all.append(inds)
        self.vals_all.append(vals)
        return

    def run(self):
        """
        Run GP-based Bayesian optimization loop
        """
        for i in range(self.exploration_steps):
            self.single_step(i)
            if self.save_checkpoints:
                self.save_results()
        self.save_results()
        if self.verbose:
            print("\nExploration completed")
        return

    def save_results(self, *args):
        """
        Save indermediary and/or final results
        """
        try:
            filename = args[0]
        except IndexError:
            filename = self.filename
        results = {}
        results['gp_pred'] = self.gp_predictions
        results['func_val'] = self.target_func_vals
        results['inds_all'] = np.array(self.indices_all)
        results['vals_all'] = np.array(self.vals_all)
        np.save(filename+".npy", results)
